"""Utils for testing"""

from inspect import Parameter, Signature
from itertools import product
from typing import (
    List,
    Any,
    Union,
    Callable,
    Iterator,
    Tuple,
    Optional,
    Iterable,
    Mapping,
)
from inspect import signature

from i2.signatures import KO, PK, PO, VP, VK, var_param_kinds
from i2.signatures import _empty
from i2.signatures import ParamsAble, Sig, ensure_param, SignatureAble

ParameterAble = Union[int, Parameter, str]
ParamsAble_ = Union[ParamsAble, str, List[int]]


def _is_valid_arg_for_sig(x):
    return (
        isinstance(x, (Callable, Signature))
        or isinstance(x, str)
        and x.startswith('(')
        and x.endswith(')')
    )


def generate_params(params: ParamsAble_):
    """Generate inspect.Parameter instances quickly.

    Example: Generate params solely from a list of their kinds

    >>> str(Sig(generate_params([0, 0, 1, 1, 1, 2, 3, 4])))
    '(a00, a01, /, a12, a13, a14, *a25, a36, **a47)'

    >>> str(Sig(generate_params("00111234")))
    '(a00, a01, /, a12, a13, a14, *a25, a36, **a47)'
    """
    if _is_valid_arg_for_sig(params):
        # generate params from a callable or signature
        yield from Sig(params).params
    else:
        for i, spec in enumerate(params):
            if isinstance(spec, int):
                kind = spec
                yield Parameter(f'a{kind}{i}', kind=kind)
            elif isinstance(spec, Parameter):
                param = spec
                yield param
            elif isinstance(spec, str) and spec.isnumeric():
                kind = int(spec)
                yield Parameter(f'a{kind}{i}', kind=kind)
            else:
                try:
                    yield ensure_param(spec)
                except Exception:
                    raise TypeError(
                        f"Don't know how to handle this type of obj: {spec}"
                    )


def params_to_arg_name_and_val(params: ParamsAble_):
    """Generate a `{argname: argval, ...}` dictionary from an iterable of params.

    >>> assert dict(params_to_arg_name_and_val(generate_params("00111234"))) == {
    ...     "a00": 0,
    ...     "a01": 1,
    ...     "a12": 2,
    ...     "a13": 3,
    ...     "a14": 4,
    ...     "a25": (5, -5),
    ...     "a36": 6,
    ...     "a47": {"a47": 7, "a47_": -7},
    ... }
    """
    params = generate_params(params)
    for i, param in enumerate(params):
        if param.kind == Parameter.VAR_POSITIONAL:
            val = (i, -i)
        elif param.kind == Parameter.VAR_KEYWORD:
            val = {param.name: i, param.name + '_': -i}
        else:
            val = i
        yield (param.name, val)


def inject_defaults(params: ParamsAble_, defaults: dict):
    """Yields params with defaults ({argname: default_val,...}) edited.

    >>> assert (
    ...     str(
    ...         Sig(
    ...             inject_defaults(
    ...                 generate_params("00111234"), defaults={"a14": 40, "a36": 60}
    ...             )
    ...         )
    ...     )
    ...     == "(a00, a01, /, a12, a13, a14=40, *a25, a36=60, **a47)"
    ... )
    """
    for param in generate_params(params):
        if param.name in defaults:
            yield param.replace(default=defaults[param.name])
        else:
            yield param


def _str_of_call_args(_call_kwargs: dict):
    return ', '.join(f'{k}={v}' for k, v in _call_kwargs.items())


def _params_to_name(params):
    return 'f' + ''.join(str(int(p.kind)) for p in params)


def mk_func_from_params(
    params: ParamsAble = '00111234',
    *,
    defaults=None,
    name=None,
    callback: Callable[[dict], Any] = _str_of_call_args,
):
    """Make a function (that actually returns something based on args) from params.

    See Also: ``sig_to_func``

    :param params: params (arguments) of the function (can be expressed in many ways!)
    :param defaults: Optional {argname: default,...} dict to inject defaults
    :param name: Optional name to give the function
    :param callback: The function defining what the function actually does.
        Must be a function taking a single dict input encapsulating the all arguments.
        The default will return a string representation of this dict.
    :return: A function with the specified params, returning a string of it's (call) args

    There's many ways you can express the `params` input.
    Any of the ways understood by the `signatures.ensure_params` function, for one;
    plus a few more.

    One nice way to express the params is through an actual function.
    Note that the code of the function isn't even looked out.
    Only it's signature is taken into consideration.
    The returned function will have the same signature.
    Instead, the callback function will be acalled on the infered _call_kwargs
    dict of {argname: argval} pairs.
    The default callaback is a string exhibiting these argname/argval pairs.

    >>> f = mk_func_from_params(lambda x, /, y, *, z: None)
    >>> print(f"{f.__name__}{Sig(f)}")
    f(x, /, y, *, z)
    >>> f(1, 2, z=3)
    'x=1, y=2, z=3'
    >>> f(1, y=2, z=3)
    'x=1, y=2, z=3'
    >>> f = mk_func_from_params(lambda x, /, y=42, *, z='ZZZ': None)
    >>> print(f"{f.__name__}{Sig(f)}")
    f(x, /, y=42, *, z='ZZZ')
    >>> f(3.14)
    'x=3.14, y=42, z=ZZZ'

    If you're not interested in having that level of control, but are just
    interested in the number and kinds of the arguments, you can specify only that;
    a sequence of kinds.
    These must be a non-decreasing sequence of integers between
    0 and 4 inclusive. These integers represent kinds of parameters.
    See https://docs.python.org/3/library/inspect.html#inspect.Parameter.kind
    to see what each integer value means.
    You can also specify this integer sequence as a single string, as shown below.

    >>> f = mk_func_from_params(params="00111234")
    >>> str(Sig(f))
    '(a00, a01, /, a12, a13, a14, *a25, a36, **a47)'
    >>> f(0, 1, 2, 3, 4, 5, -5, a36=6, a47={"a47": 7, "a47_": -7})
    "a00=0, a01=1, a12=2, a13=3, a14=4, a25=(5, -5), a36=6, a47={'a47': 7, 'a47_': -7}"
    >>> f(0, 1, 2, a13=3, a14=4, a36=6)
    'a00=0, a01=1, a12=2, a13=3, a14=4, a25=(), a36=6, a47={}'

    What just happened?
    Well, `params="00111234"` was transformed to `params=[0, 0, 1, 1, 1, 2, 3, 4]`,
    which was transformed to a list of the same size, using

    Now, if you really want full control over those params, you can specify them
    completely using the `inspect.Parameter` class.
    You can also decide what level of control you want, and mix and match all kinds of
    specifications, as below.

    >>> from inspect import Parameter
    >>> f = mk_func_from_params([
    ...     0,
    ...     'blah',
    ...     Parameter(name='hello',
    ...               kind=Parameter.POSITIONAL_OR_KEYWORD,
    ...               default='world')
    ... ])
    >>> print(f"{f.__name__}{Sig(f)}")
    f(a00, /, blah, hello='world')
    >>> assert f(11, 22) == 'a00=11, blah=22, hello=world'

    """
    params = generate_params(params)
    params = inject_defaults(params, defaults=defaults or {})
    sig = Sig(params)

    @sig
    def arg_str_func(*args, **kwargs):
        _call_kwargs = sig.kwargs_from_args_and_kwargs(
            args, kwargs, apply_defaults=True
        )
        return callback(_call_kwargs)

    arg_str_func.__name__ = name or _params_to_name(params)

    return arg_str_func


def _sig_to_str_of_call_args_code_str(sig: Sig):
    return (
        'return ' + 'f"' + _str_of_call_args({p: f'{{{p}}}' for p in sig.names}) + '"'
    )


def _is_simple_expression(code_lines):
    if len(code_lines) == 1:
        line = code_lines[0].strip()
        if not (
            line in {'pass', '...'}
            or line.startswith('return')
            or (line.startswith('"') and line.endswith("'"))
        ):
            return True
    else:
        return False


def sig_to_func(
    sig: ParamsAble = '00111234',
    code_lines: Union[
        str, Iterable, Callable[[Sig], str]
    ] = _sig_to_str_of_call_args_code_str,
    *,
    name: Optional[str] = None,
    globals: Optional[dict] = None,
    locals: Optional[Mapping] = None,
):
    """
    Make a function from a signature

    See Also: ``mk_func_from_params``

    More information: https://github.com/i2mint/i2/issues/34

    :param sig: Signature (or something that can be made into a signature) the func
    should have
    :param name: Name the function should have
    :param code_lines: The code lines

    The ``globals`` must be a dictionary and ``locals`` can be any mapping, defaulting
    to the current globals and locals. If only ``globals`` is given, ``locals``
    defaults to it.

    >>> sig = Sig('(a, /, b=2)')
    >>> f = sig_to_func(sig)
    >>> str(Sig(f))
    '(a, /, b=2)'
    >>> f(2, 3)
    'a=2, b=3'
    >>> f.__name__
    'f01'

    See how the function was given a name automatically? This name was created by
    appending the kind number
    (see https://docs.python.org/3/library/inspect.html#inspect.Parameter.kind)
    to 'f'.

    Let's demo how we can

    - use kind numbers (see `inspect` module) to specify the signature

    - give an explicit name to the function

    - specify ``locals()`` so that ``sig_to_func`` can insert the new function there
    (which makes it picklable, for instances)

    >>> _ = sig_to_func('012', name='foo', locals=locals())
    >>> # and now `foo` is in local name space, and has signature:
    >>> str(Sig(foo))
    '(a00, /, a11, *a22)'
    >>> foo(1,2,3,4,5)
    'a00=1, a11=2, a22=(3, 4, 5)'

    In the examples above, the function body was was created from the input ``sig``
    through the ``_sig_to_str_of_call_args_code_str`` function which outputs a string
    formed from the input argument names and values.

    But you can specify your own function. This function should take a ``Sig`` object
    and return the string or lines (iterable of strings) of the function's body.
    You can also specify a string or lines directly.

    >>> g = sig_to_func('(x, y=2)', 'pass')
    >>> assert g(2, y=3) is None
    >>> h = sig_to_func('(x, y=2)', 'return x * y')
    >>> h(3)
    6
    >>> h(x=10, y=3)
    30

    You can even omit the return instruction if, as in ``lambda`` functions, the body is
    a simple expression (not `pass`, `...`, or starting something with quotes).

    >>> h = sig_to_func('(x, y=2)', 'x / y')
    >>> h(10, 5)
    2.0

    """
    sig = Sig(generate_params(sig))
    name = name or _params_to_name(sig.params)
    if callable(code_lines):
        code_lines = code_lines(sig)  # call the function on sig to get lines
    if isinstance(code_lines, str):
        code_lines = code_lines.split('\n')
    if _is_simple_expression(code_lines):
        # If code_lines has only one line and it seems it's an expression, prepend return
        code_lines = [f'return {code_lines[0]}']
    code_string = '\n\t'.join(code_lines)
    func_def_string = f'def {name}{sig}:\n\t{code_string}'
    _locals = locals or {}
    exec(func_def_string, globals, _locals)
    return _locals[name]


def _all_prefixes(x: Iterable):
    """
    >>> list(_all_prefixes([1,2,3]))
    [(), (1,), (1, 2), (1, 2, 3)]
    >>> list(map(dict, _all_prefixes({'a': 1, 'b': 2}.items())))
    [{}, {'a': 1}, {'a': 1, 'b': 2}]
    """
    x = tuple(x)
    for i in range(len(x) + 1):
        yield x[:i]


def _args_kwargs_combinations(args, kwargs):
    """

    >>> assert list(_args_kwargs_combinations((1,2), {'a': 3, 'b': 4})) == [
    ...     ((), {}),
    ...     ((), {'a': 3}),
    ...     ((), {'a': 3, 'b': 4}),
    ...     ((1,), {}),
    ...     ((1,), {'a': 3}),
    ...     ((1,), {'a': 3, 'b': 4}),
    ...     ((1, 2), {}),
    ...     ((1, 2), {'a': 3}),
    ...     ((1, 2), {'a': 3, 'b': 4})
    ... ]

    """
    for args_prefix, kwargs_items_prefix in product(
        _all_prefixes(args), _all_prefixes(kwargs.items())
    ):
        yield args_prefix, dict(kwargs_items_prefix)


def variadic_type(sig, variadics):
    var_kinds = [sig.kinds[param] for param in variadics]
    if not variadics:
        return 'no_var'
    if VP in var_kinds and VK not in var_kinds:
        return 'vp_only'
    if VK in var_kinds and VP not in var_kinds:
        return 'vk_only'
    else:
        return 'vp_vk'


def create_variadic_source(sig, variadics, dflt_source):
    var_type = variadic_type(sig, variadics)
    if var_type == 'no_var':
        result = ((), {})
    elif var_type == 'vp_only':
        result = (dflt_source[0], {})
    elif var_type == 'vk_only':
        result = ((), dflt_source[1])
    elif var_type == 'vp_vk':
        result = dflt_source
    return result


# TODO: Get rid of ignore_variadics once using code is refactored
def sig_to_inputs(
    sig: SignatureAble,
    argument_vals: Optional[Iterable] = None,
    *,
    variadics_source: Tuple[tuple, dict] = (
        ('args1', 'args2'),
        {'kwargs1': 'kwargs1_val'},
    ),
) -> Iterator[Tuple[tuple, dict]]:
    """Generate all kind-valid (arg, kwargs) input combinations for a function with a
    given signature ``sig``, with argument values taken from the ``argument_vals``

    :param sig: A signature or anything that ``i2.Sig`` can use to create one (e.g.
        function, string, list of dicts etc.)
    :param argument_vals: An interable from which the argument values will be drawn.
        Defaults to ``list(range(n_args))``.
    :param variadics_source: The ``(tuple, dict)`` pair that will be used to extract
        variadics' inputs.
    :return: A generator of ``(args: tuple, kwargs: dict)`` pairs

    >>> assert list(
    ...     sig_to_inputs(lambda a, b, /, c, d, *, e, f: None)
    ... ) == [
    ...     ((0, 1), {'c': 2, 'd': 3, 'e': 4, 'f': 5}),
    ...     ((0, 1, 2), {'d': 3, 'e': 4, 'f': 5}),
    ...     ((0, 1, 2, 3), {'e': 4, 'f': 5})
    ... ]

    >>> list(sig_to_inputs(Sig('(a, *args, b, **kwargs)')))  # doctest: +NORMALIZE_WHITESPACE
    [((), {'a': 0, 'b': 1}),
     ((), {'a': 0, 'b': 1, 'kwargs1': 'kwargs1_val'}),
     (('args1',), {'a': 0, 'b': 1}),
     (('args1',), {'a': 0, 'b': 1, 'kwargs1': 'kwargs1_val'}),
     (('args1', 'args2'), {'a': 0, 'b': 1}),
     (('args1', 'args2'), {'a': 0, 'b': 1, 'kwargs1': 'kwargs1_val'}),
     ((0,), {'b': 1}),
     ((0,), {'b': 1, 'kwargs1': 'kwargs1_val'}),
     ((0, 'args1'), {'b': 1}),
     ((0, 'args1'), {'b': 1, 'kwargs1': 'kwargs1_val'}),
     ((0, 'args1', 'args2'), {'b': 1}),
     ((0, 'args1', 'args2'), {'b': 1, 'kwargs1': 'kwargs1_val'})]

    Tip: To ignore variadics all together, you can specify ``variadics_source=((), {})``.
    """
    sig = Sig(sig)
    init_sig = sig
    already_yielded = []
    variadics = [param for param, kind in sig.kinds.items() if kind in var_param_kinds]
    if variadics:
        for v in variadics:
            sig -= v
        variadics_source = create_variadic_source(
            init_sig, variadics, dflt_source=variadics_source
        )
        var_args, var_kwargs = variadics_source
        for args, kwargs in sig_to_inputs(sig, argument_vals):
            for _args, _kwargs in _args_kwargs_combinations(var_args, var_kwargs):
                yield args + _args, dict(kwargs, **_kwargs)
    else:
        for sub_sig in _get_sub_sigs_from_default_values(sig):
            po, pk, ko = _get_non_variadic_kind_counts(sub_sig)
            for args, kwargs_vals in _sig_to_inputs(
                po, pk, ko, argument_vals=argument_vals
            ):
                input_ = (
                    tuple(args),
                    {k: v for k, v in zip(sub_sig.names[len(args) :], kwargs_vals)},
                )
                if input_ not in already_yielded:
                    yield input_
                    already_yielded.append(input_)


def _sig_to_inputs(po=0, pk=0, ko=0, argument_vals: Optional[Iterable] = None):
    """

    >>> list(_sig_to_inputs(2,2,2))
    [([0, 1], [2, 3, 4, 5]), ([0, 1, 2], [3, 4, 5]), ([0, 1, 2, 3], [4, 5])]

    :param po: Number of POSITION_ONLY args in signature.
    :param pk: Number of POSITION_OR_KEYWORD args in signature.
    :param ko: Number of KEYWORD_ONLY args in signature.
    :param argument_vals: An interable from which the argument values will be drawn.
        Defaults to ``list(range(n_args))``.
    :return: A generator of ``(vals_for_args: tuple, vals_for_kwargs)`` pairs
    """
    if argument_vals is None:
        argument_vals = list(range(po + pk + ko))
    else:
        argument_vals = list(argument_vals)
    for n_args_from_pk in range(pk + 1):
        yield argument_vals[: (po + n_args_from_pk)], argument_vals[
            (po + n_args_from_pk) :
        ]


def _get_sub_sigs_from_default_values(sig: Sig) -> Iterator[Sig]:
    """Generate all the signatures compatible with the given signature
    by ignoring the arguments that have default values.
    >>> sig = Sig('(a=0, /, b=0, *args, c=0, **kwargs)')
    >>> assert [str(s) for s in _get_sub_sigs_from_default_values(sig)] == [
    ...     '(*args, **kwargs)',
    ...     '(a=0, /, *args, **kwargs)',
    ...     '(*args, b=0, **kwargs)',
    ...     '(a=0, /, b=0, *args, **kwargs)',
    ...     '(*args, c=0, **kwargs)',
    ...     '(a=0, /, *args, c=0, **kwargs)',
    ...     '(*args, b=0, c=0, **kwargs)',
    ...     '(a=0, /, b=0, *args, c=0, **kwargs)'
    ... ]
    """

    def internal_get_sub_sigs(sig):
        kos = [n for n in sig.names_of_kind[KO] if n in sig.defaults]
        for ko in reversed(kos):
            _sig = sig - ko
            yield from internal_get_sub_sigs(_sig)

        pks = [n for n in sig.names_of_kind[PK] if n in sig.defaults]
        for i, pk in reversed(list(enumerate(pks))):
            _sig = sig - pk
            pks_to_transform_to_ko = pks[i + 1 :]
            params = [
                p.replace(kind=KO) if p.name in pks_to_transform_to_ko else p
                for p in _sig.params
            ]
            params.sort(key=lambda p: p.kind)
            _sig = Sig(params)
            yield from internal_get_sub_sigs(_sig)

        pos = [n for n in sig.names_of_kind[PO] if n in sig.defaults]
        if pos:
            _sig = sig - pos[-1]
            params = [p.replace(kind=KO) if p.kind == PK else p for p in _sig.params]
            params.sort(key=lambda p: p.kind)
            _sig = Sig(params)
            yield from internal_get_sub_sigs(_sig)

        if sig not in already_yielded:
            yield sig
            already_yielded.add(sig)

    already_yielded = set()
    yield from internal_get_sub_sigs(sig)


def _get_non_variadic_kind_counts(sig: Sig):
    po = pk = ko = 0
    for kind in sig.kinds.values():
        po += kind == sig.POSITIONAL_ONLY
        pk += kind == sig.POSITIONAL_OR_KEYWORD
        ko += kind == sig.KEYWORD_ONLY
    return po, pk, ko


def mk_func_inputs_for_params(params: ParamsAble_, param_to_input):
    pass


# ---------------------------------------------------------------------------------------
# Tools to analyze compatibility between signature and function call

from i2 import Pipe
import re

_signature_msg_patterns = [
    'keyword arguments$',
    'invalid keyword argument',
    'expected at most',
    'keyword argument',
    'got some positional\-only arguments passed as keyword arguments',
    'no signature found',
]

_signature_msg_regex = re.compile('|'.join(map('({})'.format, _signature_msg_patterns)))
is_signature_msg = Pipe(_signature_msg_regex.search, bool)


def _is_signature_error(
    error_obj,
    signature_error_types=(TypeError, ValueError),
    is_signature_msg=is_signature_msg,
):
    if isinstance(error_obj, signature_error_types):
        error_msg = str(error_obj)
        return is_signature_msg(error_msg)
    return False


def call_and_return_error(func, /, *args, **kwargs):
    try:
        func(*args, **kwargs)
        return None
    except BaseException as error_obj:
        return error_obj


def on_error_return_none(func, /, *args, **kwargs):
    try:
        return func(*args, **kwargs)
    except BaseException as error_obj:
        return None


call_raises_signature_error = Pipe(call_and_return_error, _is_signature_error)

call_raises_signature_error.__doc__ = '''
>>> call_raises_signature_error(lambda x, /, y: None, 1, y=2)
False
>>> call_raises_signature_error(lambda x, /, y: None, x=1, y=2)
True
'''

# Yes, I too see that this can be made into yet another Pipe!
def function_is_compatible_with_signature(func, sig):
    """
    Runs through all combinations of positional and keyword arguments,

    >>> function_is_compatible_with_signature(hasattr, Sig(lambda obj, name: ...))
    False
    >>> function_is_compatible_with_signature(hasattr, Sig(lambda obj, name, /: ...))
    True
    """

    def _call_raises_sig_error():
        for args, kwargs in sig_to_inputs(sig):
            yield call_raises_signature_error(func, *args, **kwargs)

    return not any(_call_raises_sig_error())


def builtin_objects():
    for name in dir(__builtins__):
        yield getattr(__builtins__, name)


def builtin_signatureless_callables():
    """
    A generator of builtin callables that don't have signatures.
    """
    for obj in builtin_objects():
        if callable(obj) and call_raises_signature_error(signature, obj):
            yield obj


# ---------------------------------------------------------------------------------------

#
# @mk_func_from_params.register
# def mk_func_from_params(params: Iterable[int], defaults=None, name=None):
#     """
#
#     :param kinds:
#     :param defaults:
#     :return:
#
#     Make a sequence of kinds (must be a non-decreasing sequence of integers between
#     0 and 4 inclusive. These integers represent kinds of parameters.
#     See https://docs.python.org/3/library/inspect.html#inspect.Parameter.kind
#     to see what each integer value means.
#
#     >>> kinds = list(map(int, "00111234"))
#     >>> kinds
#     [0, 0, 1, 1, 1, 2, 3, 4]
#
#     Note: `kinds_to_arg_str_func` also works directly with strings such as "00111234".
#
#     Now tell `kinds_to_arg_str_func` to make a function with those kinds.
#
#     >>> f = kinds_to_arg_str_func(kinds)
#     >>> str(Sig(f))
#     '(a00, a01, /, a12, a13, a14, *a25, a36, **a47)'
#     >>> f(0, 1, 2, 3, 4, 5, -5, a36=6, a47={"a47": 7, "a47_": -7})
#     "a00=0, a01=1, a12=2, a13=3, a14=4, a25=(5, -5), a36=6, a47={'a47': {'a47': 7, 'a47_': -7}}"
#     >>> f(0, 1, 2, a13=3, a14=4, a36=6)
#     'a00=0, a01=1, a12=2, a13=3, a14=4, a25=(), a36=6, a47={}'
#     """
#     kinds = params
#     params = inject_defaults(generate_params(kinds), defaults=defaults or {})
#     name = name or "f" + "".join(map(str, kinds))
#
#     return mk_func_from_params(params, defaults, name)

#
#
# @mk_func_from_params.register
# def _(kinds: str):
#     """
#     >>> f = kinds_to_arg_str_func("00111234")
#     >>> str(Sig(f))
#     '(a00, a01, /, a12, a13, a14, *a25, a36, **a47)'
#     >>> f(0, 1, 2, 3, 4, 5, -5, a36=6, a47={"a47": 7, "a47_": -7})
#     "a00=0, a01=1, a12=2, a13=3, a14=4, a25=(5, -5), a36=6, a47={'a47': {'a47': 7, 'a47_': -7}}"
#     >>> f(0, 1, 2, a13=3, a14=4, a36=6)
#     'a00=0, a01=1, a12=2, a13=3, a14=4, a25=(), a36=6, a47={}'
#
#     """
#     return mk_func_from_params(_kinds_str_to_int_list(kinds))
#
#
# f = mk_func_from_params("00111234")
# assert str(Sig(f)) == "(a00, a01, /, a12, a13, a14, *a25, a36, **a47)"
# assert (
#     f(0, 1, 2, 3, 4, 5, -5, a36=6, a47={"a47": 7, "a47_": -7})
#     == "a00=0, a01=1, a12=2, a13=3, a14=4, a25=(5, -5), a36=6, a47={'a47': {'a47': 7, 'a47_': -7}}"
# )


empty = _empty

mappingproxy = type(Signature().parameters)


def trace_call(func, local_vars, name=None):
    if name is None:
        name = func.__name__
    return (
        f'{name}('
        + ', '.join(f'{argname}={local_vars[argname]}' for argname in Sig(func).names)
        + ')'
    )


# class KeywordArg(dict):
#     """Just to mark a dict as a keyword argument"""
#
#
# def _separate_pk_arguments_into_positional_and_keyword(pka):
#     args = []
#     kwargs = {}
#     pka_iter = iter(pka)
#     for a in pka_iter:
#         if not isinstance(a, KeywordArg):
#             args.append(a)
#         else:
#             kwargs.update(dict(a))
#     for a in pka_iter:
#         kwargs.update(dict(a))
#
#     return args, kwargs



"""Types"""

from typing import NewType, Optional, Iterable, Protocol, Any


def new_type(
    name,
    tp,
    doc: Optional[str] = None,
    aka: Optional[Iterable] = None,
    assign_to_globals=False,
):
    """
    Make a new type with (optional) doc and (optional) aka, set of var names it often appears as

    Args:
        name: Name to give the variable
        tp: type (see typing module)
        doc: Optional string to put in __doc__ attribute
        aka: Optional set (or any iterable) to put in _aka attribute,
            meant to list names the variables of this type often appear as.

    Returns: None

    >>> from typing import Any, Union, List
    >>> MyType = new_type('MyType', int)
    >>> type(MyType)
    <class 'function'>
    >>> Key = new_type('Key', Any, aka=['key', 'k'])
    >>> sorted(Key._aka)
    ['k', 'key']
    >>> Val = new_type('Val', Union[int, float, List[Union[int, float]]], doc="A number or list of numbers.")
    >>> Val.__doc__
    'A number or list of numbers.'
    """
    new_tp = NewType(name, tp)
    if doc is not None:
        setattr(new_tp, '__doc__', doc)
    if aka is not None:
        setattr(new_tp, '_aka', set(aka))
    if assign_to_globals:
        globals()[
            name
        ] = new_tp  # not sure how kosher this is... Should only use at top level of module, for sure!
    return new_tp


class HasAttrs:
    """
    Make a protocol to express the existence of specific attributes.

    >>> SizedAndAppendable = HasAttrs["__len__", "append"]
    >>> assert isinstance([1, 2, 3], SizedAndAppendable)  # lists have both a length and an append
    >>> assert not isinstance((1, 2, 3), SizedAndAppendable)  # tuples don't have an append

    [Python Protocols](https://www.python.org/dev/peps/pep-0544/) are a way to be able to do
    "behavior typing" (my bad terminology).
    Basically, if you want your static analyzer
    (the swingles in your IDE, or linter validation process...)
    to check if you're manipulating the expected types, except the types
    (classes, subclasses, ABCs, abstract classes...) are too restrictive (they are!),
    you can use Protocols to fill the gap.

    Except writing them can sometimes be verbose.

    With HasAttrs you can have the basic "does it have these attributes" cases covered.

    >>> assert isinstance(dict(), HasAttrs["items"])
    >>> assert not isinstance(list(), HasAttrs["items"])
    >>> assert not isinstance(dict(), HasAttrs["append"])
    >>>
    >>> class A:
    ...     prop = 2
    ...
    ...     def method(self):
    ...         pass
    >>>
    >>> a = A()
    >>> assert isinstance(a, HasAttrs["method"])
    >>> assert isinstance(a, HasAttrs["method", "prop"])
    >>> assert not isinstance(a, HasAttrs["method", "prop", "this_attr_does_not_exist"])

    """

    def __class_getitem__(self, attr_names):
        if isinstance(attr_names, str):
            attr_names = [attr_names]
        assert all(map(str.isidentifier, attr_names)), (
            f"The following are not valid python 'identifiers' "
            f"{', '.join(a for a in attr_names if not a.isidentifier())}"
        )

        annotations = {attr: Any for attr in attr_names}

        class HasAttrs(Protocol):
            __annotations__ = annotations

            # def __repr__(self):  # TODO: This is for the instance, need it for the class
            #     return "HasAttrs[{', '.join(annotations)}]"

        return HasAttrs

        # return type(
        #     "HasAttrs",
        #     (Protocol,),
        #     {"__annotations__": {attr: Any for attr in attr_names}},
        # )


# TODO: Complete scary hack: Find another way (see uses)
def is_a_new_type(typ):
    return (
        callable(typ)
        and getattr(typ, '__qualname__', '').startswith('NewType')
        and hasattr(typ, '__supertype__')
    )


def typ_name(typ):
    if is_a_new_type(typ):
        return typ.__name__
    else:
        return typ._name


def is_callable_kind(typ):
    """
    >>> from typing import Callable, Tuple
    >>> is_callable_kind(Callable)
    True
    >>> is_callable_kind(Callable[[int, float], str])
    True
    >>> is_callable_kind(Tuple[int, float, str])
    False
    """
    if is_a_new_type(typ):
        return is_callable_kind(typ.__supertype__)
    return typ_name(typ) == 'Callable'
    # Also possible: typ.mro()[0] == __import__('collections.abc').Callable


def input_and_output_types(typ):
    """

    :param typ:
    :return:


    >>> from typing import Callable, Tuple
    >>> input_types, output_type = input_and_output_types(Callable[[float, int], str])
    >>> assert input_types == (float, int) and output_type == str
    >>> input_types, output_type = input_and_output_types(Callable[[], str])
    >>> assert input_types == () and output_type == str

    But will fail if `typ` isn't a `Callable`:

    >>> input_and_output_types(Tuple[float, int, str])  # doctest: +NORMALIZE_WHITESPACE
    Traceback (most recent call last):
      ...
    AssertionError: Is not a typing.Callable kind: typing.Tuple[float, int, str]

    Will also fail if `typ` is a Callable but not "parametrized".

    >>> input_and_output_types(Callable)  # doctest: +NORMALIZE_WHITESPACE
    Traceback (most recent call last):
      ...
    AssertionError: Can only be used on a Callable[[...],...] kind: typing.Callable

    """
    if is_a_new_type(typ):
        return input_and_output_types(typ.__supertype__)
    assert is_callable_kind(typ), f'Is not a typing.Callable kind: {typ}'
    assert (
        len(typ.__args__) > 0
    ), f'Can only be used on a Callable[[...],...] kind: {typ}'
    return typ.__args__[:-1], typ.__args__[-1]


def dot_string_of_callable_typ(typ):
    input_types, output_type = input_and_output_types(typ)
    return (
        ','.join(map(typ_name, input_types))
        + f' -> {typ_name(typ)} -> '
        + typ_name(output_type)
    )


def dot_strings_of_callable_types(*typs, func_shape='box'):
    for typ in typs:
        yield dot_string_of_callable_typ(typ)
        yield f'{typ_name(typ)} [shape="{func_shape}"]'



"""
Tools to specify functions through trees and forests.

Whaaa?!?

Well, you see, often -- especially when writing transformers -- you have a series of
if/then conditions nested into eachother, in code, where it gets ugly and un-reusable.

This module explores ways to objectivy this: That is, to give us the means to create
such nested conditions in a way that we can define the parts as reusable operable
components.

Think of the relationship between the for loop (code) and the iterator (object), along
with iterator tools (itertools).
This is what we're trying to explore, but for if/then conditions.

I said explore. Some more work is needed here to make it easily usable.

Let's look at an example involving the three main actors of our play.
Each of these are ``Iterable`` and ``Callable`` (``Generator`` to be precise).

- ``CondNode``: implements the if/then (no else) logic
- ``FinalNode``: Final -- yields (both with call and iter) it's single `.val` attribute.
- ``RoutingForest``: An Iterable of ``CondNode``

>>> import inspect
>>>
>>> def could_be_int(obj):
...     if isinstance(obj, int):
...         b = True
...     else:
...         try:
...             int(obj)
...             b = True
...         except ValueError:
...             b = False
...     if b:
...         print(f'{inspect.currentframe().f_code.co_name}')
...     return b
...
>>> def could_be_float(obj):
...     if isinstance(obj, float):
...         b = True
...     else:
...         try:
...             float(obj)
...             b = True
...         except ValueError:
...             b = False
...     if b:
...         print(f'{inspect.currentframe().f_code.co_name}')
...     return b
...
>>> print(
...     could_be_int(30),
...     could_be_int(30.3),
...     could_be_int('30.2'),
...     could_be_int('nope'),
... )
could_be_int
could_be_int
True True False False
>>> print(
...     could_be_float(30),
...     could_be_float(30.3),
...     could_be_float('30.2'),
...     could_be_float('nope'),
... )
could_be_float
could_be_float
could_be_float
True True True False
>>> assert could_be_int('30.2') is False
>>> assert could_be_float('30.2') is True
could_be_float
>>>
>>> st = RoutingForest(
...     [
...         CondNode(
...             cond=could_be_int,
...             then=RoutingForest(
...                 [
...                     CondNode(
...                         cond=lambda x: int(x) >= 10,
...                         then=FinalNode('More than a digit'),
...                     ),
...                     CondNode(
...                         cond=lambda x: (int(x) % 2) == 1,
...                         then=FinalNode("That's odd!"),
...                     ),
...                 ]
...             ),
...         ),
...         CondNode(cond=could_be_float, then=FinalNode('could be seen as a float')),
...     ]
... )
>>> assert list(st('nothing I can do with that')) == []
>>> assert list(st(8)) == ['could be seen as a float']
could_be_int
could_be_float
>>> assert list(st(9)) == ["That's odd!", 'could be seen as a float']
could_be_int
could_be_float
>>> assert list(st(10)) == ['More than a digit', 'could be seen as a float']
could_be_int
could_be_float
>>> assert list(st(11)) == [
...     'More than a digit',
...     "That's odd!",
...     'could be seen as a float',
... ]
could_be_int
could_be_float
>>>
>>> print(
...     '### RoutingForest ########################################################################################'
... )
### RoutingForest ########################################################################################
>>> rf = RoutingForest(
...     [
...         SwitchCaseNode(
...             switch=lambda x: x % 5,
...             cases={0: FinalNode('zero_mod_5'), 1: FinalNode('one_mod_5')},
...             default=FinalNode('default_mod_5'),
...         ),
...         SwitchCaseNode(
...             switch=lambda x: x % 2,
...             cases={0: FinalNode('even'), 1: FinalNode('odd')},
...             default=FinalNode('that is not an int'),
...         ),
...     ]
... )
>>>
>>> assert list(rf(5)) == ['zero_mod_5', 'odd']
>>> assert list(rf(6)) == ['one_mod_5', 'even']
>>> assert list(rf(7)) == ['default_mod_5', 'odd']
>>> assert list(rf(8)) == ['default_mod_5', 'even']
>>> assert list(rf(10)) == ['zero_mod_5', 'even']
>>>

"""
from itertools import chain
from dataclasses import dataclass
from typing import Any, Iterable, Callable, Mapping, Tuple
from i2.util import Literal


# TODO: Think a bit harder about this mini-language
# TODO: Right now use has to explicitly declare final nodes. Can do better.
def _default_mini_lang(x):
    # TODO: One is really tempted to use CondNode here to define this, right?
    if isinstance(x, (CondNode, RoutingForest, SwitchCaseNode)):
        return x
    elif isinstance(x, Literal):
        return x.get_val()
    elif isinstance(x, tuple):
        if len(x) == 2:
            return CondNode(*map(_default_mini_lang, x))
        elif len(x) == 3:
            return SwitchCaseNode(*map(_default_mini_lang, x))
        else:
            raise ValueError(
                f'If a tuple, element must be a `(cond, then)` pair, '
                f'or a (switch, case, default) triple, or a `Literal`. Was: {x}'
            )
    elif isinstance(x, dict):
        keys = set(x)
        if 2 <= len(keys) <= 3:
            if {'cond', 'then'}.issubset(keys):
                return CondNode(**{k: _default_mini_lang(v) for k, v in x.items()})
            elif {'switch', 'case'}.issubset(keys) and keys == {
                'switch',
                'case',
                'default',
            }:
                return SwitchCaseNode(
                    **{k: _default_mini_lang(v) for k, v in x.items()}
                )
            else:
                raise ValueError(
                    "keys should be 'switch', 'case' and optionally 'default'. "
                    f'Were: {keys}'
                )
        else:
            raise ValueError(
                f"A non-Literal dict must have keys 'cond' and 'then' "
                f"or 'switch' and 'case' (and optionally 'default'). Was: {x}"
            )
    elif isinstance(x, list):
        return RoutingForest(cond_nodes=list(map(_default_mini_lang, x)))
    else:
        return x


class RoutingNode:
    """A RoutingNode instance needs to be callable on a single object, yielding an iterable or a final value"""

    def __call__(self, obj):
        raise NotImplementedError('You should implement this.')

    @staticmethod
    def from_object(x, mini_lang=_default_mini_lang):
        return mini_lang(x)


@dataclass
class FinalNode(RoutingNode):
    """A RoutingNode that is final.
    It yields (both with call and iter) it's single `.val` attribute."""

    val: Any

    def __call__(self, obj=None):
        yield self.val

    def __iter__(self):
        yield self.val

    # def __getstate__(self):
    #     return {'val': self.val}


# TODO: Add tooling to merge validation into `then` functions/values
@dataclass
class CondNode(RoutingNode):
    """A RoutingNode that implements the if/then (no else) logic"""

    cond: Callable[[Any], bool]
    then: Any

    def __call__(self, obj):
        if self.cond(obj):
            yield from self.then(obj)

    def __iter__(self):
        yield from self.then


@dataclass
class RoutingForest(RoutingNode):
    """

    >>> rf = RoutingForest([
    ...     CondNode(cond=lambda x: isinstance(x, int),
    ...              then=RoutingForest([
    ...                  CondNode(cond=lambda x: int(x) >= 10, then=FinalNode('More than a digit')),
    ...                  CondNode(cond=lambda x: (int(x) % 2) == 1, then=FinalNode("That's odd!"))])
    ...             ),
    ...     CondNode(cond=lambda x: isinstance(x, (int, float)),
    ...              then=FinalNode('could be seen as a float')),
    ... ])
    >>> assert list(rf('nothing I can do with that')) == []
    >>> assert list(rf(8)) == ['could be seen as a float']
    >>> assert list(rf(9)) == ["That's odd!", 'could be seen as a float']
    >>> assert list(rf(10)) == ['More than a digit', 'could be seen as a float']
    >>> assert list(rf(11)) == ['More than a digit', "That's odd!", 'could be seen as a float']
    """

    cond_nodes: Iterable

    def __post_init__(self):
        self.cond_nodes = list(map(self.from_object, self.cond_nodes))

    def __call__(self, obj):
        yield from chain(*(cond_node(obj) for cond_node in self.cond_nodes))
        # for cond_node in self.cond_nodes:
        #     yield from cond_node(obj)

    def __iter__(self):
        yield from chain(*self.cond_nodes)


FeatCondThens = Iterable[Tuple[Callable, Callable]]


@dataclass
class FeatCondNode(RoutingNode):
    """A RoutingNode that yields multiple routes, one for each of several conditions met,
    where the condition is computed implements computes a feature of the obj and
    according to an iterable of conditions on the feature"""

    feat: Callable
    feat_cond_thens: FeatCondThens

    def __call__(self, obj):
        feature = self.feat(obj)
        for cond, then in self.feat_cond_thens:
            if cond(feature):
                yield from then(obj)

    def __iter__(self):
        yield from chain(*self.feat_cond_thens.values())


NoDefault = type('NoDefault', (object,), {})
NO_DFLT = NoDefault()


@dataclass
class SwitchCaseNode(RoutingNode):
    """A RoutingNode that implements the switch/case/else logic.
    It's just a specialization (enhanced with a "default" option) of the FeatCondNode class to a situation
    where the cond function of feat_cond_thens is equality, therefore the routing can be
    implemented with a {value_to_compare_to_feature: then_node} map.
    :param switch: A function returning the feature of an object we want to switch on
    :param cases: The mapping from feature to RoutingNode that should be yield for that feature.
        Often is a dict, but only requirement is that it implements the cases.get(val, default) method.
    :param default: Default RoutingNode to yield if no

    >>> rf = RoutingForest([
    ...     SwitchCaseNode(switch=lambda x: x % 5,
    ...                    cases={0: FinalNode('zero_mod_5'), 1: FinalNode('one_mod_5')},
    ...                    default=FinalNode('default_mod_5')),
    ...     SwitchCaseNode(switch=lambda x: x % 2,
    ...                    cases={0: FinalNode('even'), 1: FinalNode('odd')},
    ...                    default=FinalNode('that is not an int')),
    ... ])
    >>>
    >>> assert(list(rf(5)) == ['zero_mod_5', 'odd'])
    >>> assert(list(rf(6)) == ['one_mod_5', 'even'])
    >>> assert(list(rf(7)) == ['default_mod_5', 'odd'])
    >>> assert(list(rf(8)) == ['default_mod_5', 'even'])
    >>> assert(list(rf(10)) == ['zero_mod_5', 'even'])
    """

    switch: Callable
    cases: Mapping
    default: Any = NO_DFLT

    def __call__(self, obj):
        feature = self.switch(obj)
        if self.default is NO_DFLT:
            yield from self.cases.get(feature)(obj)
        else:
            yield from self.cases.get(feature, self.default)(obj)

    def __iter__(self):
        yield from chain(*self.cases.values())
        if self.default:
            yield self.default


def wrap_leafs_with_final_node(x):
    for xx in x:
        if isinstance(xx, RoutingNode):
            yield xx
        else:
            yield FinalNode(xx)


def test_routing_forest():
    def could_be_int(obj):
        if isinstance(obj, int):
            b = True
        else:
            try:
                int(obj)
                b = True
            except ValueError:
                b = False
        # if b:
        #     print(f'{inspect.currentframe().f_code.co_name}')
        return b

    def could_be_float(obj):
        if isinstance(obj, float):
            b = True
        else:
            try:
                float(obj)
                b = True
            except ValueError:
                b = False
        # if b:
        #     print(f'{inspect.currentframe().f_code.co_name}')
        return b

    assert could_be_int(30)
    assert could_be_int(30.3)
    assert not could_be_int('30.2')
    assert not could_be_int('nope')

    assert could_be_float(30)
    assert could_be_float(30.3)
    assert could_be_float('30.2')
    assert not could_be_float('nope')

    assert could_be_int('30.2') is False
    assert could_be_float('30.2') is True

    st = RoutingForest(
        [
            CondNode(
                cond=could_be_int,
                then=RoutingForest(
                    [
                        CondNode(
                            cond=lambda x: int(x) >= 10,
                            then=FinalNode('More than a digit'),
                        ),
                        CondNode(
                            cond=lambda x: (int(x) % 2) == 1,
                            then=FinalNode("That's odd!"),
                        ),
                    ]
                ),
            ),
            CondNode(cond=could_be_float, then=FinalNode('could be seen as a float')),
        ]
    )
    assert list(st('nothing I can do with that')) == []
    assert list(st(8)) == ['could be seen as a float']
    assert list(st(9)) == ["That's odd!", 'could be seen as a float']
    assert list(st(10)) == ['More than a digit', 'could be seen as a float']
    assert list(st(11)) == [
        'More than a digit',
        "That's odd!",
        'could be seen as a float',
    ]

    rf = RoutingForest(
        [
            SwitchCaseNode(
                switch=lambda x: x % 5,
                cases={0: FinalNode('zero_mod_5'), 1: FinalNode('one_mod_5')},
                default=FinalNode('default_mod_5'),
            ),
            SwitchCaseNode(
                switch=lambda x: x % 2,
                cases={0: FinalNode('even'), 1: FinalNode('odd')},
                default=FinalNode('that is not an int'),
            ),
        ]
    )

    assert list(rf(5)) == ['zero_mod_5', 'odd']
    assert list(rf(6)) == ['one_mod_5', 'even']
    assert list(rf(7)) == ['default_mod_5', 'odd']
    assert list(rf(8)) == ['default_mod_5', 'even']
    assert list(rf(10)) == ['zero_mod_5', 'even']

    # testing default mini-language #####################################################

    st2 = RoutingNode.from_object(
        [  # RoutingForest
            (  # CondNode
                could_be_int,
                [  # RoutingForest
                    (lambda x: int(x) >= 10, FinalNode('More than a digit')),
                    (lambda x: (int(x) % 2) == 1, FinalNode("That's odd!")),
                ],
            ),
            (could_be_float, FinalNode('could be seen as a float')),
        ]
    )

    assert list(st2('nothing I can do with that')) == []
    assert list(st2(8)) == ['could be seen as a float']
    assert list(st2(9)) == ["That's odd!", 'could be seen as a float']
    assert list(st2(10)) == ['More than a digit', 'could be seen as a float']
    assert list(st2(11)) == [
        'More than a digit',
        "That's odd!",
        'could be seen as a float',
    ]

    rf2 = RoutingForest(
        [
            (
                lambda x: x % 5,
                Literal({0: FinalNode('zero_mod_5'), 1: FinalNode('one_mod_5')}),
                FinalNode('default_mod_5'),
            ),
            (
                lambda x: x % 2,
                Literal({0: FinalNode('even'), 1: FinalNode('odd')}),
                FinalNode('that is not an int'),
            ),
        ]
    )

    assert list(rf2(5)) == ['zero_mod_5', 'odd']
    assert list(rf2(6)) == ['one_mod_5', 'even']
    assert list(rf2(7)) == ['default_mod_5', 'odd']
    assert list(rf2(8)) == ['default_mod_5', 'even']
    assert list(rf2(10)) == ['zero_mod_5', 'even']



"""Testing wrapper"""

from i2.wrapper import wrap, mk_ingress_from_name_mapper


def _test_ingress(a, b: str, c='hi'):
    return (a + len(b) % 2,), dict(string=f'{c} {b}')


def _test_func(times, string):
    return times * string


def test_wrap():
    import pickle
    from inspect import signature

    func = _test_func

    # Just wrapping the func gives you a sort of copy of the func.
    wrapped_func = wrap(func)  # no transformations
    assert wrapped_func(2, 'co') == 'coco' == func(2, 'co')

    # If you give the wrap an ingress function
    ingress = _test_ingress
    wrapped_func = wrap(func, ingress=ingress)
    # it will use it to (1) determine the signature of the wrapped_func
    assert (
        str(signature(wrapped_func)) == "(a, b: str, c='hi')"
    )  # "(a, b: str, c='hi')"
    # and (2) to map inputs
    assert wrapped_func(2, 'world! ', 'Hi') == 'Hi world! Hi world! Hi world! '

    # An egress function can be used to transform outputs
    wrapped_func = wrap(func, egress=len)
    assert wrapped_func(2, 'co') == 4 == len('coco') == len(func(2, 'co'))

    # Both ingress and egress can be used in combination
    wrapped_func = wrap(func, ingress=ingress, egress=len)
    assert (
        wrapped_func(2, 'world! ', 'Hi') == 30 == len('Hi world! Hi world! Hi world! ')
    )

    # A wrapped function is pickle-able (unlike the usual way decorators are written)

    unpickled_wrapped_func = pickle.loads(pickle.dumps(wrapped_func))
    assert (
        unpickled_wrapped_func(2, 'world! ', 'Hi')
        == 30
        == len('Hi world! Hi world! Hi world! ')
    )


def _test_foo(a, b: int, c=7):
    return a + b * c


def _test_bar(a, /, b: int, *, c=7):
    return a + b * c


def test_mk_ingress_from_name_mapper():
    import pickle
    from inspect import signature

    foo = _test_foo
    # Define the mapping (keys are inner and values are outer names)
    name_mapper = dict(a='aa', c='cc')
    # Make an ingress function with that mapping
    ingress = mk_ingress_from_name_mapper(foo, name_mapper)
    # Use the ingress function to wrap a function
    wrapped_foo = wrap(foo, ingress=ingress)
    # See that the signature of the wrapped func uses the mapped arg names
    assert (
        str(signature(wrapped_foo)) == str(signature(ingress)) == '(aa, b: int, cc=7)'
    )
    # And that wrapped function does compute correctly
    assert (
        foo(1, 2, c=4)
        == foo(a=1, b=2, c=4)
        == wrapped_foo(aa=1, b=2, cc=4)
        == wrapped_foo(1, 2, cc=4)
    )
    # The ingress function returns args and kwargs for wrapped function
    assert ingress('i was called aa', b='i am b', cc=42) == (
        (),
        {'a': 'i was called aa', 'b': 'i am b', 'c': 42},
    )
    # See above that the args is empty. That will be the case most of the time.
    # Keyword arguments will be favored when there's a choice. If wrapped
    # function uses position-only arguments though, ingress will have to use them
    bar = _test_bar
    assert str(signature(bar)) == '(a, /, b: int, *, c=7)'
    ingress_for_bar = mk_ingress_from_name_mapper(bar, name_mapper)
    assert ingress_for_bar('i was called aa', b='i am b', cc=42) == (
        ('i was called aa',),
        {'b': 'i am b', 'c': 42},
    )
    wrapped_bar = wrap(bar, ingress=ingress_for_bar)
    assert (
        bar(1, 2, c=4)
        == bar(1, b=2, c=4)
        == wrapped_bar(1, b=2, cc=4)
        == wrapped_bar(1, 2, cc=4)
    )

    # Note that though bar had a positional only and a keyword only argument,
    # we are (by default) free of argument kind constraints in the wrapped function:
    # We can can use a positional args on `cc` and keyword args on `aa`
    assert str(signature(wrapped_bar)) == '(aa, b: int, cc=7)'
    assert wrapped_bar(1, 2, 4) == wrapped_bar(aa=1, b=2, cc=4)

    # If you want to conserve the argument kinds of the wrapped function, you can
    # specify this with `conserve_kind=True`:
    ingress_for_bar = mk_ingress_from_name_mapper(bar, name_mapper, conserve_kind=True)
    wrapped_bar = wrap(bar, ingress=ingress_for_bar)
    assert str(signature(wrapped_bar)) == '(aa, /, b: int, *, cc=7)'

    # A wrapped function is pickle-able (unlike the usual way decorators are written)
    unpickled_wrapped_foo = pickle.loads(pickle.dumps(wrapped_foo))
    assert (
        str(signature(unpickled_wrapped_foo))
        == str(signature(ingress))
        == '(aa, b: int, cc=7)'
    )
    assert (
        foo(1, 2, c=4)
        == foo(a=1, b=2, c=4)
        == unpickled_wrapped_foo(aa=1, b=2, cc=4)
        == unpickled_wrapped_foo(1, 2, cc=4)
    )


def test_arg_val_converter():
    from i2.wrapper import arg_val_converter
    from i2.tests.objects_for_testing import formula1, times_2, plus_1

    assert formula1(4, 3, 2, z=1) == 14
    assert formula1(4, 3 * 2, 2, z=1 + 1) == 400

    # See that "transparently" converting the function doesn't change anything

    formula2 = arg_val_converter(formula1)
    assert formula2(4, 3, 2, z=1) == 14 == formula1(4, 3, 2, z=1)

    # But now if we ask to convert x and z...

    formula2 = arg_val_converter(formula1, x=times_2, z=plus_1)
    assert formula2(4, 3, 2, z=1) == 400 == formula1(4, 3 * 2, 2, z=1 + 1)

    from i2.tests.test_util import unpickled_func_still_works

    assert unpickled_func_still_works(formula1, 4, 3, 2, z=1)
    assert unpickled_func_still_works(formula2, 4, 3, 2, z=1)


from i2.wrapper import Wrapx


def test_wrapx():
    from inspect import signature

    # Test that an trivial Wrapx instance (i.e. no ingress, caller, or egress
    # modifications) gives us an object behaving like the wrapped function.

    # TODO: Make it work with param kinds: e.g. func(x: int, *, y=1) -> int:
    def func(x: int, y=1) -> int:
        return x + y

    wrapped_func = Wrapx(func)
    assert (
        str(signature(wrapped_func)) == '(x: int, y=1) -> int' == str(signature(func))
    )

    # Test egress that has a single param z

    def func(x, y):
        return x + y

    def egress(v, *, z):
        return v * z

    wrapped_func = Wrapx(func, egress=egress)

    assert func(1, 2) == 3

    # TODO: should be '(x, y=1, *, z)' --> Need to work on the merge for this.
    assert str(signature(wrapped_func)) == '(x, y, z)'
    assert wrapped_func(1, 2, z=3) == 9 == func(1, 2) * 3

    # A more realistic application: Saving outputs to a specific location on output

    def func(x, y):
        return x + y

    def save_on_output_egress(v, *, k, s):
        s[k] = v
        return v

    save_on_output = Wrapx(func, egress=save_on_output_egress)
    # TODO: should be `(x, y, *, k, s)` --> Need to work on the merge for this.
    assert str(signature(save_on_output)) == '(x, y, k, s)'

    store = dict()
    save_on_output(1, 2, k='save_here', s=store)
    assert save_on_output(1, 2, k='save_here', s=store) == 3 == func(1, 2)
    assert store == {'save_here': 3}

    # Trying out a caller: Here, we want to wrap the function so it will apply to an
    # iterable of inputs, returning a list of results

    def func(x, y=2):
        return x + y

    def iterize(func, args, kwargs):
        first_arg_val = next(iter(kwargs.values()))
        return list(map(func, first_arg_val))

    iterized_func = Wrapx(func, caller=iterize)

    assert iterized_func([1, 2, 3, 4]) == [3, 4, 5, 6]

    # The same as above, except generalized to allow other variables (here ``y``) to
    # be input as well

    from functools import partial

    def func(x, y):
        return x + y

    def iterize_first_arg(func, args, kwargs):
        first_arg_name = next(iter(kwargs))
        remaining_kwargs = {k: v for k, v in kwargs.items() if k != first_arg_name}
        return list(map(partial(func, **remaining_kwargs), kwargs[first_arg_name]))

    iterized_func = Wrapx(func, caller=iterize_first_arg)

    assert iterized_func([1, 2, 3, 4], 10) == [11, 12, 13, 14]



try:
    import pytest
except ImportError as err:
    import warnings

    warnings.warn(
        "You don't seem to have pytest, so I can't use it to test. Shame. pytest is nice."
    )
    warnings.warn(f'Error was: {err}')


def func(obj):
    return obj.a + obj.b


class A:
    e = 2

    def __init__(self, a=1, b=0, c=1, d=10):
        self.a = a
        self.b = b
        self.c = c
        self.d = d

    def target_method(self, x):
        """Accesses ['a', 'b', 'c', 'e']"""
        t = func(self)  # and this function will access some attributes!
        tt = self.other_method(t)
        return x * tt / self.e

    def other_method(self, x=1):
        """Accesses ['c', 'e']"""
        w = self.c * 2  # c is accessed first
        return self.e + self.c * x - w  # and c is accessed again

    @classmethod
    def a_class_method(cls, y):
        """Accesses ['e']"""
        return cls.e + y


class B:
    x = 1
    y = 2
    b = 'i am b'
    c = 'i am not b'

    def __init__(self, a, greeting='hello', y=3):
        self.a = a
        self.greeting = greeting
        self.y = y  # instance y overwrites class y

    def greet(self, person='world'):
        """Accesses a, x, and y"""
        return str(self.a) + ' ' + person + '! ' + 'x+y=' + str(self.x + self.y)

    @property
    def z(self):
        """Accesses x and y"""
        return self.x + self.y

    def accessing_property_method(self):
        """Accesses b and z (and through z, x, and y)"""
        return len(self.b) + self.z

    def with_f_string(self):
        """Accesses greeting and x, but through f-string"""
        return f'{self.greeting} {self.x}'

    def writing_to_an_attribute(self, x_val, a_val):
        """Accesses x (class attr) and a (instance attr), but only to write in it.
        Should it be listed as "accessed".
        I say: Yes, if not too hard to do, but only if requested (what is NECESSARY) for
        the method to be computed is what is needed the most. Therefore only attrs that are getted, not setted.
        """
        self.x = x_val
        self.a = a_val


def convert_output():
    pass


def test_attrs_used_by_method():
    from i2.footprints import attrs_used_by_method

    assert attrs_used_by_method(A.target_method) == {'a', 'b', 'c', 'e'}
    assert attrs_used_by_method(A.other_method) == {'c', 'e'}
    assert attrs_used_by_method(A.a_class_method) == {'e'}

    assert attrs_used_by_method(B.greet) == {'a', 'x', 'y'}
    # assert attrs_used_by_method(B.z) == {'x', 'y'}
    # assert attrs_used_by_method(B.accessing_property_method) == {'b', 'z', 'x', 'y'}  # perhaps z should not be here?
    assert attrs_used_by_method(B.with_f_string) == {'greeting', 'x'}
    # assert attrs_used_by_method(B.writing_to_an_attribute) == {'x', 'a'}


def test_attrs_used_by_method_computation():
    from i2.footprints import attrs_used_by_method_computation

    assert attrs_used_by_method_computation(A.target_method, {}, {'x': 3}) == {
        'a',
        'b',
        'c',
        'e',
    }
    assert attrs_used_by_method_computation(A.other_method, {}) == {'c', 'e'}
    # assert attrs_used_by_method_computation(A.a_class_method, {}, {'y': 3}) == {'e'}  # fails (returns {})

    init_kws = dict(a=100)
    assert attrs_used_by_method_computation(B.greet, init_kws) == {
        'a',
        'x',
        'y',
    }
    assert attrs_used_by_method_computation(B.z, init_kws) == {'x', 'y'}
    assert attrs_used_by_method_computation(B.accessing_property_method, init_kws) == {
        'b',
        'z',
        'x',
        'y',
    }  # z or not z?
    assert attrs_used_by_method_computation(B.with_f_string, init_kws) == {
        'greeting',
        'x',
    }
    assert attrs_used_by_method_computation(
        B.writing_to_an_attribute, init_kws, dict(x_val=0, a_val=1)
    ) == {'x', 'a'}


## Order conserving
# def test_attrs_used_by_method():
#     from py2mint.footprints import attrs_used_by_method
#
#     assert attrs_used_by_method(A.target_method) == ['a', 'b', 'c', 'e']
#     assert attrs_used_by_method(A.other_method) == ['c', 'e']
#     assert attrs_used_by_method(A.a_class_method) == ['e']
#
#     assert attrs_used_by_method(B.greet) == ['a', 'x', 'y']
#     assert attrs_used_by_method(B.z) == ['x', 'y']
#     assert attrs_used_by_method(B.accessing_property_method) == ['b', 'z', 'x', 'y']  # perhaps z should not be here?
#     assert attrs_used_by_method(B.with_f_string) == ['greeting', 'x']
#     assert attrs_used_by_method(B.writing_to_an_attribute) == ['x', 'a']
#
#
# def test_attrs_used_by_method_computation():
#     from py2mint.footprints import attrs_used_by_method_computation
#
#     assert attrs_used_by_method_computation(A.target_method, {}, {'x': 3}) == ['a', 'b', 'c', 'e']
#     assert attrs_used_by_method_computation(A.other_method, {}) == ['c', 'e']
#     assert attrs_used_by_method_computation(A.a_class_method, {}) == ['e']
#
#     init_kws = dict(a=100)
#     assert attrs_used_by_method_computation(B.greet, init_kws) == ['a', 'x', 'y']
#     assert attrs_used_by_method_computation(B.z, init_kws) == ['x', 'y']
#     assert attrs_used_by_method_computation(
#         B.accessing_property_method, init_kws) == ['b', 'z', 'x', 'y']  # z or not z?
#     assert attrs_used_by_method_computation(B.with_f_string, init_kws) == ['greeting', 'x']
#     assert attrs_used_by_method_computation(
#         B.writing_to_an_attribute, init_kws, dict(x_val=0, y_val=1)) == ['x', 'a']



"""A wrapper object and tools to work with it

How the ``Wrap`` class works:

.. code-block::

          *outer_args, **outer_kwargs
                     │
                     ▼
    ┌───────────────────────────────────┐
    │              ingress              │
    └───────────────────────────────────┘
                     │
                     ▼
          *inner_args, **inner_kwargs
                     │
                     ▼
    ┌───────────────────────────────────┐
    │               func                │
    └───────────────────────────────────┘
                     │
                     ▼
                 func_output
                     │
                     ▼
    ┌───────────────────────────────────┐
    │              egress               │
    └───────────────────────────────────┘
                     │
                     ▼
                final_output


How the ``Ingress`` class (ingress templated function maker) works:

.. code-block::

          *outer_args, **outer_kwargs
                     │
                     ▼
    ┌───────────────────────────────────┐
    │          outer_sig_bind           │
    └───────────────────────────────────┘
                     │
                     ▼
              outer_all_kwargs
                     │
                     ▼
    ┌───────────────────────────────────┐
    │            kwargs_trans           │
    └───────────────────────────────────┘
                     │
                     ▼
              inner_all_kwargs
                     │
                     ▼
    ┌───────────────────────────────────┐
    │          inner_sig_bind           │
    └───────────────────────────────────┘
                     │
                     ▼
          *inner_args, **inner_kwargs

"""

from functools import wraps, partial
from inspect import Parameter, signature
from typing import (
    Mapping,
    Callable,
    Optional,
    Union,
    Iterable,
    Sequence,
    NewType,
    Dict,
    Tuple,
)

from types import MethodType

from i2.signatures import Sig, name_of_obj, KO, PK, VK, sig_to_dataclass
from i2.multi_object import Pipe
from i2.deco import double_up_as_factory

# ---------------------------------------------------------------------------------------
# Wrap

empty = Parameter.empty
OuterKwargs = dict
InnerKwargs = dict
KwargsTrans = Callable[[OuterKwargs], InnerKwargs]


def identity(x):
    """Transparent function, returning what's been input"""
    return x


def transparent_ingress(*args, **kwargs):
    """
    >>> transparent_ingress(1, 2, test=1)
    ((1, 2), {'test': 1})
    """

    return args, kwargs


def transparent_egress(output):
    """
    >>> transparent_egress('unnecessary_doctest')
    'unnecessary_doctest'
    """

    return output


class MakeFromFunc:
    """Used to indicate that an object should be made as a function of an input func"""

    def __init__(self, func_to_obj):
        self.func_to_obj = func_to_obj

    def __call__(self, func):
        return self.func_to_obj(func)


# from wrapt import decorator
# TODO: Continue factoring out Wrap and Wrapx code
class _Wrap:
    """To be used as the base of actual Wrap objects."""

    _init_args = ()
    _init_kwargs = ()

    def __init__(self, func, *args, **kwargs):
        self._init_args = (func, *args)
        self._init_kwargs = kwargs
        # wraps(func)(self) is there to copy over to self anything that func may
        # have had. It should be before anything else so it doesn't overwrite stuff
        # that we may add to self in init (like .func for example!)
        wraps(func)(self)  # TODO: should we really copy everything by default?
        if name := kwargs.get('name', None) is not None:
            self.__name__ = name
        self.func = func  # Note: overwrites self.func that wraps MAY have inserted
        self.__wrapped__ = func
        # TODO: Pros and cons analysis of pointing __wrapped__ to func. partial uses
        #  .func, but wraps looks for __wrapped__

    def __call__(self, *args, **kwargs):
        """Just forward the call to the wrapped function."""
        return self.func(*args, **kwargs)

    def __reduce__(self):
        """reduce is meant to control how things are pickled/unpickled"""
        return type(self), self._init_args, dict(self._init_kwargs)

    def __get__(self, instance, owner):
        """This method allows things to work well when we use Wrap object as method"""
        if instance is None:
            return self
        return MethodType(self, instance)

    def __repr__(self):
        # TODO: Replace i2.Wrap with dynamic (Wrap or Wrapx)
        name = getattr(self, '__name__', None) or 'Wrap'
        return f'<i2.Wrap {name}{signature(self)}>'

    # TODO: Don't know exactly what I'm doing below. Review with someone!
    def __set_name__(self, owner, name):
        """So that name of function is passed on to method when assigning to attribute
        That is, doing ``method = Wrap(func)`` in a class definition"""
        # TODO: Look into sanity of mutating the name and other ways to achieve same
        self.__name__ = name

    def __set__(self, instance, value):
        instance.__dict__[self.__name__] = value

    # To get help(instance.method) to work!
    # TODO: Does this have undesirable side effects?
    @property
    def __code__(self):
        return self.func.__code__


def _defaults_and_kwdefaults_of_func(func: Callable):
    try:
        return func.__defaults__, func.__kwdefaults__
    except AttributeError:
        # if python can't do it on it's own, you may have a special kind of callable
        # here, so try doing it through Sig, which is more flexible
        sig = Sig(func)
        return sig._defaults_, sig._kwdefaults_


class Wrap(_Wrap):
    """A function wrapper with interface modifiers.

    :param func: The wrapped function
    :param ingress: The incoming data transformer. It determines the argument properties
        (name, kind, default and annotation) as well as the actual input of the
        wrapped function.
    :param egress: The outgoing data transformer. It also takes precedence over the
        wrapped function to determine the return annotation of the ``Wrap`` instance
    :param name: Name to give the wrapper (will use wrapped func name by default)
    :return: A callable instance wrapping ``func``

    Some examples:

    >>> from inspect import signature
    >>> from i2 import Sig

    >>> def func(a, b):
    ...     return a * b

    >>> wrapped_func = wrap(func)  # no transformations: wrapped_func is the same as func
    >>> assert wrapped_func(2, 'Hi') == func(2, 'Hi') == 'HiHi'

    Modifying the first argument

    >>> def ingress(a, b):
    ...   return (2 * a, b), dict()
    >>> wrapped_func = wrap(func, ingress=ingress)  # first variable is now multiplied by 2
    >>> wrapped_func(2, 'Hi')
    'HiHiHiHi'

    Same using keyword args, we need to use tuple to represent an empty tuple

    >>> def ingress(a, b):
    ...   return tuple(), dict(a=2 * a, b=b) # Note that b MUST be present as well, or an error will be raised
    >>> wrapped_func = wrap(func, ingress=ingress)  # first variable is now multiplied by 2
    >>> wrapped_func(2, 'Hi')
    'HiHiHiHi'

    Using both args and kwargs

    >>> def ingress(a, b):
    ...   return (2 * a, ), dict(b=b)
    >>> wrapped_func = wrap(func, ingress=ingress)  # first variable is now multiplied by 2
    >>> wrapped_func(2, 'Hi')
    'HiHiHiHi'

    We can use ingress to ADD parameters to func

    >>> def ingress(a, b, c):
    ...   return (a, b + c), dict()
    >>> wrapped_func = wrap(func, ingress=ingress)
    >>> # now wrapped_func takes three arguments
    >>> wrapped_func(2, 'Hi', 'world!')
    'Hiworld!Hiworld!'

    Egress is a bit more straightforward, it simply applies to the output of the
    wrapped function. We can use ingress to ADD parameters to func

    >>> def egress(output):
    ...   return output + ' ITSME!!!'
    >>> wrapped_func = wrap(func, ingress=ingress, egress=egress)
    >>> # now wrapped_func takes three arguments
    >>> wrapped_func(2, 'Hi', 'world!')
    'Hiworld!Hiworld! ITSME!!!'

    A more involved example:

    >>> def ingress(a, b: str, c="hi"):
    ...     return (a + len(b) % 2,), dict(string=f"{c} {b}")
    ...
    >>> def func(times, string):
    ...     return times * string
    ...
    >>> wrapped_func = wrap(func, ingress=ingress)
    >>> assert wrapped_func(2, "world! ", "Hi") == "Hi world! Hi world! Hi world! "
    >>>
    >>> wrapped_func = wrap(func, egress=len)
    >>> assert wrapped_func(2, "co") == 4 == len("coco") == len(func(2, "co"))
    >>>
    >>> wrapped_func = wrap(func, ingress=ingress, egress=len)
    >>> assert (
    ...     wrapped_func(2, "world! ", "Hi")
    ...     == 30
    ...     == len("Hi world! Hi world! Hi world! ")
    ... )

    An ``ingress`` function links the interface of the wrapper to the interface of the
    wrapped func; therefore it's definition often depends on information of both,
    and for that reason, we provide the ability to specify the ingress not only
    explicitly (as in the examples above), but through a factory -- a function that
    will be called on ``func`` to produce the ingress that should be used to wrap it.

    .. seealso::

        ``wrap`` function.

    """

    def __init__(self, func, ingress=None, egress=None, *, name=None):
        super().__init__(func, ingress, egress, name=name)
        ingress_sig = Sig(func)

        if ingress is None:
            self.ingress = transparent_ingress
            self.__defaults__, self.__kwdefaults__ = _defaults_and_kwdefaults_of_func(
                func
            )
        else:

            if isinstance(ingress, MakeFromFunc):
                self.ingress = ingress  # the ingress function is
                func_to_ingress = ingress  # it's not the ingress function itself
                # ... but an ingress factory: Should make the ingress in function of func
                self.ingress = func_to_ingress(func)
            else:
                assert callable(ingress), f'Should be callable: {ingress}'
                self.ingress = ingress

            ingress_sig = Sig(self.ingress)
            self.__defaults__, self.__kwdefaults__ = _defaults_and_kwdefaults_of_func(
                self.ingress
            )

        return_annotation = empty

        if egress is None:
            self.egress = transparent_egress
        else:
            self.egress = egress
            egress_return_annotation = Sig(egress).return_annotation
            if egress_return_annotation is not Parameter.empty:
                return_annotation = egress_return_annotation

        self.__signature__ = Sig(ingress_sig, return_annotation=return_annotation)

    def __call__(self, *ingress_args, **ingress_kwargs):
        func_args, func_kwargs = self.ingress(*ingress_args, **ingress_kwargs)
        return self.egress(self.func(*func_args, **func_kwargs))


@double_up_as_factory
def wrap(
    func=None, *, ingress=None, egress=None, caller=None, name=None, dflt_wrap=Wrap
):
    """Wrap a function, optionally transforming interface, input and output.

    :param func: The wrapped function
    :param ingress: The incoming data transformer. It determines the argument properties
        (name, kind, default and annotation) as well as the actual input of the
        wrapped function.
    :param egress: The outgoing data transformer. It also takes precedence over the
        wrapped function to determine the return annotation of the ``Wrap`` instance
    :return: A callable instance wrapping ``func``

    Consider the following function.

    >>> def f(w, /, x: float = 1, y=2, *, z: int = 3):
    ...     return w + x * y ** z
    ...
    >>> assert f(0) == 8
    >>> assert f(1,2) == 17 == 1 + 2 * 2 ** 3

    See that ``f`` is restricted to use ``z`` as keyword only argument kind:

    >>> f(1, 2, 3, 4)
    Traceback (most recent call last):
      ...
    TypeError: f() takes from 1 to 3 positional arguments but 4 were given

    and ``w`` has position only argument kind:

    >>> f(w=1, x=2, y=3, z=4)
    Traceback (most recent call last):
      ...
    TypeError: f() got some positional-only arguments passed as keyword arguments: 'w'

    Say we wanted a version of this function that didn't have the argument kind
    restrinctions, where the annotation of ``x`` was ``int`` and where the default
    of ``z`` was ``10`` instead of ``3``, and doesn't have an annotation.
    We can do so using the following ingress function:

    >>> def ingress(w, x: int = 1, y: int=2, z = 10):
    ...     return (w,), dict(x=x, y=y, z=z)

    The ingress function serves two purposes:

    - Redefining the signature (i.e. the argument names, kinds, defaults,
    and annotations (not including the return annotation, which is taken care of by the
    egress argument).

    - Telling the wrapper how to get from that interface to the interface of the
    wrapped function.

    If we also wanted to add a return_annotation, we could do so via an ``egress``
    function argument:

    >>> def egress(wrapped_func_output) -> float:
    ...     return wrapped_func_output  # because here we don't want to do anything extra

    Now we can use these ingress and egress functions to get the version of ``f`` of
    our dreams:

    >>> h = wrap(f, ingress=ingress, egress=egress)

    Let's see what the signature of our new function looks like:

    >>> from inspect import signature
    >>> str(signature(h))
    '(w, x: int = 1, y: int = 2, z=10) -> float'

    Now let's see that we can actually use this new function ``h``, without the
    restrictions of argument kind, getting the same results as the wrapped ``f``,
    but with default ``z=10``.

    What we wanted (but couldn't) do with ``f``:

    >>> h(1, 2, 3, 4)  # == 1 + 2 * 3 ** 4
    163
    >>> h(w=1, x=2, y=3, z=4)
    163

    >>> assert h(0) == h(0, 1) == h(0, 1, 2) == 0 + 1 * 2 ** 10 == 2 ** 10 == 1024

    Note that ``wrap`` can also be used as a decorator "factory", for instance to
    wrap functions at definition time, and if we change ``caller`` it will automatically
    use ``Wrapx`` instead of ``Wrap`` to wrap the function.

    >>> def iterize(func, args, kwargs):
    ...     first_arg_val = next(iter(kwargs.values()))
    ...     return list(map(func, first_arg_val))
    >>>
    >>> @wrap(caller=iterize)
    ... def func(x, y=2):
    ...     return x + y
    ...
    >>> isinstance(func, Wrapx)
    True
    >>> func([1, 2, 3, 4])
    [3, 4, 5, 6]

    For more examples, see also the

    .. seealso::

        ``Wrap`` class.
        ``Wrapx`` class.

    """
    if _should_use_wrapx(func, ingress, egress, caller):
        return Wrapx(func, ingress, egress, caller=caller, name=name)
    else:
        return dflt_wrap(func, ingress, egress, name=name)


# TODO: Add conditions on egress to route to Wrapx when complex egress
def _should_use_wrapx(func, ingress, egress, caller):
    if caller is not None:
        return True
    else:
        return False


def append_empty_args(func):
    """To use to transform an ingress function that only returns kwargs to one that
    returns the normal form of ingress functions: ((), kwargs)"""

    @wraps(func)
    def _func(*args, **kwargs):
        return (), func(*args, **kwargs)

    return _func


class Ingress:
    """The Ingress class offers a template for creating ingress classes.

    Note that when writing a decorator with i2.wrapper, you're usually better off
    writing an ingress function for the purpose. As a result, your code will usually
    be less complex, easier to read, and more efficient than using the Ingress class.

    So why use the Ingress class at all? For one, because it'll take care of some common
    mechanics for you, so once you understand how to use it,
    you'll probably create a correct wrapper faster.

    Further, if you're writing a general wrapping tool (e.g. your own currying machine,
    some rule-based input casting function, etc.) then you'll find that using
    Ingres will usually with on the complexity, readability and/or efficiency front.

    >>> from i2.wrapper import Ingress, wrap
    >>> from inspect import signature
    >>> from i2.wrapper import InnerMapIngress
    >>>
    >>> # Ingress = InnerMapIngress
    >>>
    >>> def f(w, /, x: float, y=2, *, z: int = 3):
    ...     return f"(w:={w}) + (x:={x}) * (y:={y}) ** (z:={z}) == {w + x * y ** z}"
    >>>
    >>> f(0, 1)
    '(w:=0) + (x:=1) * (y:=2) ** (z:=3) == 8'

    Let’s say you wanted to dispatch this function to a command line interface,
    or a webservice where all arguments are taken from the url.
    The problem here is that this means that all incoming values will be strings
    in that case.
    Say you wanted all input values to be cast to ints. In that case you could do:

    >>> trans_all_vals_to_ints = lambda d: {k: int(v) for k, v in d.items()}
    >>>
    >>> cli_f = wrap(
    ...     f,
    ...     ingress=Ingress(signature(f), kwargs_trans=trans_all_vals_to_ints)
    ... )
    >>>
    >>> cli_f("2", "3", "4")
    '(w:=2) + (x:=3) * (y:=4) ** (z:=3) == 194'

    In a more realistic situation, you'd want to have more control over this value
    transformation.

    Say you wanted to convert to int if it's possible, try float if not,
    and just leave the string alone otherwise.

    >>> def _try_casting_to_numeric(x):
    ...     try:
    ...         return int(x)
    ...     except ValueError:
    ...         try:
    ...             return float(x)
    ...         except ValueError:
    ...             return x
    ...
    >>> def cast_numbers(d: dict):
    ...     return {k: _try_casting_to_numeric(v) for k, v in d.items()}
    >>>
    >>> cli_f = wrap(f, ingress=Ingress(signature(f), kwargs_trans=cast_numbers))
    >>>
    >>> cli_f("2", "3.14", "4")
    '(w:=2) + (x:=3.14) * (y:=4) ** (z:=3) == 202.96'

    Let's say that our values transformations are not all 1-to-1 as in the examples
    above.
    Instead, they can be

    - ``1-to-many`` (e.g. the outer 'w' is used to compute the inner ``w`` and ``x``)

    - ``many-to-1 (e.g. the outer ``x`` and ``y`` are used to compute inner ``y``)

    .. code-block::
          w   x   y   z
         / \   \ /    |
        w   x   y     z


    >>> def kwargs_trans(outer_kw):
    ...     return dict(
    ...         # e.g. 1-to-many: one outer arg (w) producing two inner args (w, and y)
    ...         w=outer_kw['w'] * 2,
    ...         x=outer_kw['w'] * 3,
    ...         # e.g. many-to-1: two outer args (x and y) producing one inner arg (y)
    ...         y=outer_kw['x'] + outer_kw['y'],
    ...         # Note that no z is mentioned: This means we're just leaving it alone
    ...     )
    ...
    >>>
    >>> ingress = Ingress(signature(f), kwargs_trans=kwargs_trans)
    >>> assert ingress(2, x=3, y=4) == ((4,), {'x': 6, 'y': 7, 'z': 3})
    >>>
    >>> wrapped_f = wrap(f, ingress=ingress)
    >>> assert wrapped_f(2, x=3, y=4) == '(w:=4) + (x:=6) * (y:=7) ** (z:=3) == 2062'


    The following is an example that involves several aspects of the ``Ingress`` class.

    >>> from i2 import Sig
    >>> def kwargs_trans(outer_kw):
    ...     return dict(
    ...         w=outer_kw['w'] * 2,
    ...         x=outer_kw['w'] * 3,
    ...         # need to pop you (inner func has no you argument)
    ...         y=outer_kw['x'] + outer_kw.pop('you'),
    ...         # Note that no z is mentioned: This means we're just leaving it alone
    ...     )
    >>>
    >>> ingress = Ingress(
    ...     inner_sig=signature(f),
    ...     kwargs_trans=kwargs_trans,
    ...     outer_sig=Sig(f).ch_names(y='you')  # need to give the outer sig a you
    ...     # You could also express it this way (though you'd lose the annotations)
    ...     # outer_sig=lambda w, /, x, you=2, *, z=3: None
    ... )
    >>> assert ingress(2, x=3, you=4) == ((4,), {'x': 6, 'y': 7, 'z': 3})
    >>>
    >>> wrapped_f = wrap(f, ingress=ingress)
    >>> assert wrapped_f(2, x=3, you=4) == '(w:=4) + (x:=6) * (y:=7) ** (z:=3) == 2062'

    A convenience method allows to do the same with the ingress instance itself:

    >>> wrapped_f = ingress.wrap(f)
    >>> assert wrapped_f(2, x=3, you=4) == '(w:=4) + (x:=6) * (y:=7) ** (z:=3) == 2062'
    """

    def __init__(
        self, inner_sig, kwargs_trans: Optional[KwargsTrans] = None, outer_sig=None
    ):
        """Init of an Ingress instance.

        :param inner_sig: Signature of the inner function the ingress is for.
            The function itself can be given and the signature will be extracted.
        :param kwargs_trans: A dict-to-dict transformation of the outer kwargs to
            the kwargs that should be input to the inner function.
            That is ``kwargs_trans`` is ``outer_kwargs -> inner_kwargs``.
            Note that though both outer and inner signatures could have those annoying
            position-only kinds, you don't have to think of that.
            The parameter kind restrictions are taken care of automatically.
        :param outer_sig: The outer signature. The signature the ingress function
            (there for the wrapped function) will have. Also serves to convert input
            (args, kwargs) to the kwargs that will be given to kwargs_trans.

        When making an Ingress function directly, one must take care that
        ``inner_sig``, ``kwargs_trans`` and ``outer_sig`` are aligned.

        Namely, 'kwargs_trans' must be able to handle outputs of
        ``outer_sig.kwargs_from_args_and_kwargs`` and itself output kwargs that
        can be handled by ``inner_sig.args_and_kwargs_from_kwargs``.

        """
        self.inner_sig = Sig(inner_sig)

        # kwargs_trans should be callable and have one required arg: a dict
        # if it's None, we'll just make it be the identity function
        if kwargs_trans is None:
            kwargs_trans = identity
        self.kwargs_trans = kwargs_trans

        # default to inner_sig = outer_sig
        if outer_sig is None:
            outer_sig = inner_sig
        self.outer_sig = Sig(outer_sig)

        self.outer_sig(self)

    def __call__(self, *ingress_args, **ingress_kwargs):
        # Get the all-keywords version of the arguments (args,kwargs->kwargs)
        func_kwargs = self.outer_sig.kwargs_from_args_and_kwargs(
            ingress_args, ingress_kwargs, apply_defaults=True
        )

        func_kwargs = dict(
            func_kwargs,  # by default, keep the func_kwargs, but
            **self.kwargs_trans(func_kwargs),  # change those that kwargs_trans desires
        )

        # Return an (args, kwargs) pair that respects the inner function's
        # argument kind restrictions.
        # Note: Originally was with (default) allow_excess=False. Changed to True to
        #       allow more flexibility in outer sig. But is this sane? Worth it?
        # TODO: Reflect on pros/cons of allow_excess=True
        return self.inner_sig.args_and_kwargs_from_kwargs(
            func_kwargs, apply_defaults=True, allow_excess=True
        )

    def __repr__(self):
        return f'Ingress signature: {signature(self)}'

    def wrap(self, func: Callable, egress=None, *, name=None) -> Wrap:
        """Convenience method to wrap a function with the instance ingress.
        ``ingress.wrap(func,...)`` equivalent to ``Wrap(func, ingress, ...)``
        """
        return Wrap(func, ingress=self, egress=egress, name=name)

    @classmethod
    def name_map(cls, wrapped, **old_to_new_name):
        """Change argument names.

        >>> def f(w, /, x: float, y=2, *, z: int = 3):
        ...     return f"(w:={w}) + (x:={x}) * (y:={y}) ** (z:={z}) == {w + x * y ** z}"
        >>> ingress = Ingress.name_map(f, w='DoubleYou', z='Zee')
        >>> ingress
        Ingress signature: (DoubleYou, /, x: float, y=2, *, Zee: int = 3)
        >>> wrapped_f = ingress.wrap(f)
        >>> wrapped_f(1, 2, y=3, Zee=4)
        '(w:=1) + (x:=2) * (y:=3) ** (z:=4) == 163'

        """
        new_to_old_name = {v: k for k, v in old_to_new_name.items()}
        assert len(new_to_old_name) == len(
            old_to_new_name
        ), f'Inversion is not possible since {old_to_new_name=} has duplicate values.'
        return cls(
            wrapped,
            partial(Pipe(items_with_mapped_keys, dict), key_mapper=new_to_old_name),
            Sig(wrapped).ch_names(**old_to_new_name),
        )

    #     @classmethod
    #     def defaults(cls, wrapped, **defaults):
    #         """"""
    #
    #     @classmethod
    #     def order(cls, wrapped, arg_order):
    #         """"""
    #
    #     @classmethod
    #     def factory(cls, wrapped, **func_for_name):
    #         """"""


def items_with_mapped_keys(d: dict, key_mapper):
    """Transform dict keys. More precisely yield (new_k,v) pairs from a key mapper dict.

    :param d: src dict
    :param key_mapper: {old_name: new_name, ...} mapping
    :return: generator of (new_name, value) pairs

    Often used in conjunction with dict:

    >>> dict(items_with_mapped_keys(
    ...     {'a': 1, 'b': 2, 'c': 3, 'd': 4},
    ...     {'a': 'Ay', 'd': 'Dee'})
    ... )
    {'Ay': 1, 'b': 2, 'c': 3, 'Dee': 4}

    """
    for k, v in d.items():
        # key_mapper.get(k, k) will give the new key name if present,
        # else will use the old
        yield key_mapper.get(k, k), v


def invert_map(d: dict):
    new_d = {v: k for k, v in d.items()}
    if len(new_d) == len(d):
        return new_d
    else:
        raise ValueError(f'There are duplicate keys so I can invert map: {d}')


from i2.signatures import parameter_to_dict


def parameters_to_dict(parameters):
    return {name: parameter_to_dict(param) for name, param in parameters.items()}


def _handle_ingress_class_inputs(
    inner_sig,
    kwargs_trans: Optional[KwargsTrans],
    outer_sig,
    *,
    _allow_reordering=False,
):
    inner_sig = Sig(inner_sig)

    # kwargs_trans should be callable and have one required arg: a dict
    # if it's None, we'll just make it be the identity function
    if kwargs_trans is None:
        kwargs_trans = identity

    # default to inner_sig = outer_sig
    if outer_sig is None:
        outer_sig = inner_sig  # if nothing specified, want same outer and inner sigs
    elif isinstance(outer_sig, Mapping):
        changes_for_name = outer_sig  # it's a dict of modifications of the inner sig
        outer_sig = inner_sig.modified(
            _allow_reordering=_allow_reordering, **changes_for_name
        )
    else:
        outer_sig = Sig(outer_sig)

    return inner_sig, kwargs_trans, outer_sig


# TODO: See what this adds over ``Ingress`` class. Consider merging or reusing.
class InnerMapIngress:
    """A class to help build ingresses systematically by mapping the inner signature.

    *Systematically*, i.e. "according to a fixed plan/system" is what it's about here.
    As we'll see below, if you need to write a particular adapter for a specific case,
    you probably should do by writing an actual ingress function directly.
    In cases where you might want to apply a same logic to wrap many functions,
    you may want to fix that wrapping logic: ``InnerMapIngress`` provides one
    way to do this.

    :param inner_sig: The signature of the wrapped function.
    :param kwargs_trans: A dict-to-dict transformation of the outer kwargs to
        the kwargs that should be input to the inner function.
        That is ``kwargs_trans`` is ``outer_kwargs -> inner_kwargs``.
        Note that though both outer and inner signatures could have those annoying
        position-only kinds, you don't have to think of that.
        The parameter kind restrictions are taken care of automatically.
    :param _allow_reordering: Whether we want to allow reordering of variables
    :param in_to_out_sig_changes: The ``inner_name=dict_of_changes_for_that_name``
    pairs, the ``dict_of_changes_for_that_name`` is a ``dict`` with keys being valid
    ``inspect.Parameter``

    Consider the following function that has a position only, a keyword only,
    two arguments with annotations, and three with a default.

    >>> def f(w, /, x: float = 1, y=2, *, z: int = 3):
    ...     return w + x * y ** z

    Say we wanted a version of this function

    - that didn't have the argument kind restrinctions (all POSITION_OR_KEYWORD),

    - where the annotation of ``x`` was changed ``int`` and the default removed

    - where ``y`` was named ``you`` instead, and has an annotation (``int``).

    - where the default of ``z`` was ``10`` instead of ``3``, and doesn't have an
    annotation.

    In order to get a version of this function we wanted (more lenient kinds,
    with some annotations and a default change), we can use the ingress function:

    >>> def directly_defined_ingress(w, x: int, you: int=2, z = 10):
    ...     return (w,), dict(x=x, y=you, z=z)


    When we need to wrap a specific function in a specific way, defining an
    ingress function  this way is usually the simplest way.
    But in some cases we need to build the ingress function using some predefined
    rule/protocol to make applying the rule/protocol systematic.

    For those cases, ``InnerMapIngress`` comes in handy.

    With ``InnerMapIngress`` we'd build our ingress function like this:

    >>> from inspect import Parameter, signature
    >>> PK = Parameter.POSITIONAL_OR_KEYWORD
    >>> empty = Parameter.empty
    >>> ingress = InnerMapIngress(
    ...     f,
    ...     # change kind to PK:
    ...     w=dict(kind=PK),
    ...     # change annotation of x from float to int and remove default
    ...     x=dict(annotation=int, default=empty),
    ...     # rename y to you and add annotation int:
    ...     y=dict(name='you', annotation=int),
    ...     # change kind to PK, default to 10, and remove annotation:
    ...     z=dict(kind=PK, default=10, annotation=empty),
    ... )

    Note:

    - Only the changes we wish to make to the parameters are mentioned.
        You could also define the parameters explicitly by simply listing all three
        of the dimensions (kind, annotation, and default)

    - Three? But a ``Parameter`` object has four; what about the name?
        Indeed, you can use name as well, more on that later.

    - Note that in order to specify that you want no default, or no annotation,
        you cannot use ``None`` since ``None`` is both a valid default and a valid
        annotation; So instead you need to use ``Parameter.empty`` (conveniently
        assigned to a constant named ``empty`` in the ``wrapping`` module.

    Now see that all arguments are ``POSITIONAL_OR_KEYWORD``, ``x`` and ``y`` are
    ``int``, and default of ``z`` is 10:

    >>> assert (
    ...     str(signature(ingress))
    ...     == str(signature(directly_defined_ingress))
    ...     == '(w, x: int, you: int = 2, z=10)'
    ... )

    Additionally, ``ingress`` function does it's job of dispatching the right args
    and kwargs to the target function:

    >>> assert (
    ...     ingress(0,1,2,3)
    ...     == directly_defined_ingress(0,1,2,3)
    ...     == ((0,), {'x': 1, 'y': 2, 'z': 3})
    ... )


    """

    def __init__(
        self,
        inner_sig,
        kwargs_trans: Optional[KwargsTrans] = None,
        *,
        _allow_reordering=False,
        **changes_for_name,
    ):
        self.inner_sig = Sig(inner_sig)

        self.outer_sig = self.inner_sig.modified(
            _allow_reordering=_allow_reordering, **changes_for_name
        )

        # kwargs_trans should be callable and have one required arg: a dict
        # if it's None, we'll just make it be the identity function
        if kwargs_trans is None:
            kwargs_trans = identity
        self.kwargs_trans = kwargs_trans

        outer_name_for_inner_name = {
            inner_name: change['name']
            for inner_name, change in changes_for_name.items()
            if 'name' in change
        }
        self.inner_name_for_outer_name = invert_map(outer_name_for_inner_name)
        self.outer_sig(self)

    def __call__(self, *ingress_args, **ingress_kwargs):
        # Get the all-keywords version of the arguments (args,kwargs->kwargs)
        func_kwargs = self.outer_sig.kwargs_from_args_and_kwargs(
            ingress_args, ingress_kwargs, apply_defaults=True
        )

        # Modify the keys of func_kwargs so they reflect the inner signature's names
        # That is, map outer names to inner names.
        func_kwargs = dict(
            items_with_mapped_keys(func_kwargs, self.inner_name_for_outer_name)
        )
        func_kwargs = dict(
            func_kwargs,  # by default, keep the func_kwargs, but
            **self.kwargs_trans(func_kwargs),  # change those that kwargs_trans desires
        )

        # Return an (args, kwargs) pair the respects the inner function's
        # argument kind restrictions.
        return self.inner_sig.args_and_kwargs_from_kwargs(func_kwargs)

    @classmethod
    def from_signature(cls, inner_sig, outer_sig, _allow_reordering=False):
        """

        :param inner_sig:
        :param outer_sig:
        :param _allow_reordering:
        :return:

        Say we wanted to get a version of the function:

        >>> def f(w, /, x: float = 1, y=2, *, z: int = 3):
        ...     return w + x * y ** z

        That was equivalent to (note the kind, default and annotation differences):

        >>> def g(w, x=1, y=2, z=10):
        ...     return w + x * y ** z


        >>> ingress = InnerMapIngress.from_signature(
        ... f, outer_sig=lambda w, x=1, y=2, z=10: None
        ... )
        >>> Sig(ingress)
        <Sig (w, x=1, y=2, z=10)>
        >>>
        >>>
        >>>
        >>> h = wrap(f, ingress=InnerMapIngress.from_signature(f, g))
        >>> assert h(0) == g(0) == 1024 == 0 + 1 * 2 ** 10
        >>> assert h(1,2) == g(1,2) == 2049 == 1 + 2 * 2 ** 10
        >>> assert h(1,2,3,4) == g(1,2,3,4) == 1 + 2 * 3 ** 4
        >>>
        >>> assert h(w=1,x=2,y=3,z=4) == g(1,2,3,4) == 1 + 2 * 3 ** 4
        """
        outer_sig = Sig(outer_sig)
        return cls(
            inner_sig,
            _allow_reordering=_allow_reordering,
            **parameters_to_dict(outer_sig.parameters),
        )


# TODO: Fits global pattern -- merge
class ArgNameMappingIngress:
    def __init__(self, inner_sig, *, conserve_kind=False, **outer_name_for_inner_name):
        self.inner_sig = Sig(inner_sig)
        self.outer_sig = self.inner_sig.ch_names(**outer_name_for_inner_name)
        if conserve_kind is not True:
            self.outer_sig = self.outer_sig.ch_kinds_to_position_or_keyword()
        self.inner_name_for_outer_name = invert_map(outer_name_for_inner_name)
        self.outer_sig(self)

    def __call__(self, *ingress_args, **ingress_kwargs):
        # Get the all-keywords version of the arguments (args,kwargs->kwargs)
        func_kwargs = self.outer_sig.kwargs_from_args_and_kwargs(
            ingress_args, ingress_kwargs
        )
        # Modify the keys of func_kwargs so they reflect the inner signature's names
        # That is, map outer names to inner names.
        func_kwargs = dict(
            items_with_mapped_keys(func_kwargs, self.inner_name_for_outer_name)
        )

        # Return an (args,kwargs) pair the respects the inner function's
        # argument kind restrictions.
        return self.inner_sig.args_and_kwargs_from_kwargs(func_kwargs)


def mk_ingress_from_name_mapper(func, name_mapper: Mapping, *, conserve_kind=False):
    return ArgNameMappingIngress(func, conserve_kind=conserve_kind, **name_mapper)


# TODO: Replace with i2.wrapper instead
# TODO: Make sure VARIADICs are handled properly, or error raised if present


def apply_func_on_cond(func, cond, k, v):
    if cond(k, v):
        return func(v)
    else:
        return v


def modify_dict_on_cond(d, cond, func):
    return {k: apply_func_on_cond(func, cond, k, v) for k, v in d.items()}


def convert_to_PK(kinds):
    return {name: PK for name in kinds}


def convert_VK_to_KO(kinds):
    cond = lambda k, v: v == VK
    func = lambda v: KO

    return modify_dict_on_cond(kinds, cond, func)


def nice_kinds(func, kinds_modifier=convert_to_PK):
    """Wraps the func, changing the argument kinds according to kinds_modifier.
    The default behaviour is to change all kinds to POSITIONAL_OR_KEYWORD kinds.
    The original purpose of this function is to remove argument-kind restriction
    annoyances when doing functional manipulations such as:

    >>> from functools import partial
    >>> isinstance_of_str = partial(isinstance, class_or_tuple=str)
    >>> isinstance_of_str('I am a string')
    Traceback (most recent call last):
      ...
    TypeError: isinstance() takes no keyword arguments

    Here, instead, we can just get a kinder version of the function and do what we
    want to do:

    >>> _isinstance = nice_kinds(isinstance)
    >>> isinstance_of_str = partial(_isinstance, class_or_tuple=str)
    >>> isinstance_of_str('I am a string')
    True
    >>> isinstance_of_str(42)
    False

    See also: ``i2.signatures.all_pk_signature``

    """
    from i2 import Sig, call_somewhat_forgivingly

    sig = Sig(func)
    kinds_modif = kinds_modifier(sig.kinds)

    sig = sig.ch_kinds(**kinds_modif)

    # sig = sig.ch_kinds(**{name: Sig.POSITIONAL_OR_KEYWORD for name in sig.names})

    @wraps(func)
    def _func(*args, **kwargs):
        return call_somewhat_forgivingly(func, args, kwargs, enforce_sig=sig)

    _func.__signature__ = sig
    return _func


def wrap_from_sig(func, new_sig):
    from i2 import call_somewhat_forgivingly

    @wraps(func)
    def _func(*args, **kwargs):
        return call_somewhat_forgivingly(func, args, kwargs, enforce_sig=new_sig)

    _func.__signature__ = new_sig

    return _func


# ---------------------------------------------------------------------------------------
# wrap tools


@double_up_as_factory
def ch_names(func=None, **old_to_new_name):
    """Change the argument names of a function.

    >>> def f(w, /, x: float, y=2, *, z: int = 3):
    ...     return f"(w:={w}) + (x:={x}) * (y:={y}) ** (z:={z}) == {w + x * y ** z}"
    >>> wrapped_f = ch_names(f, w='DoubleYou', z='Zee')
    >>> wrapped_f
    <i2.Wrap f(DoubleYou, /, x: float, y=2, *, Zee: int = 3)>
    >>> wrapped_f(1, 2, y=3, Zee=4)
    '(w:=1) + (x:=2) * (y:=3) ** (z:=4) == 163'

    Can also be used as a factory:
    >>> @ch_names(a='alpha', g='gamma')
    ... def foo(a, b, g=1):
    ...     return a + b * g
    >>> foo(alpha=1, b=2, gamma=3)
    7
    """
    return Ingress.name_map(func, **old_to_new_name).wrap(func)


map_names = ch_names  # back-compatibility alias


def include_exclude_ingress_factory(func, include=None, exclude=None):
    """A pattern underlying any ingress that takes a subset of parameters (possibly
    reordering them).

    For example: Keep only required arguments, or reorder params to be able to
    partialize #3 (without having to partialize #1 and #2)

    Note: A more general version would allow include and exclude to be expressed as
    functions that apply to one or several properties of the params (name, kind, default,
    annotation).
    """
    sig = Sig(func)
    if isinstance(include, str):
        include = include.split()
    if isinstance(exclude, str):
        exclude = exclude.split()
    exclude = exclude or set()
    include = [x for x in (include or sig.names) if x not in exclude]

    return Ingress(inner_sig=sig, outer_sig=sig[include])


@double_up_as_factory
def include_exclude(func=None, *, include=None, exclude=None):
    """Reorder and/or remove parameters.

    >>> def foo(a, b, c='C', d='D'):
    ...     print(f"{a=},{b=},{c=},{d=}")
    >>> bar = include_exclude(foo, include='b a', exclude='c d')

    The signature of ``bar`` has only ``b`` and ``a``, in that order:

    >>> from inspect import signature
    >>> str(signature(bar))
    '(b, a)'

    But the function still works and does the same thing:

    >>> bar('B', 'A')
    a='A',b='B',c='C',d='D'

    """
    return wrap(func, ingress=include_exclude_ingress_factory(func, include, exclude))


@double_up_as_factory
def rm_params(func=None, *, params_to_remove=()):
    """Get a function with some parameters removed.

    >>> from inspect import signature
    >>> def func(x, y=1, z=2):
    ...     return x + y * z
    >>>
    >>> f = rm_params(func, params_to_remove='z')
    >>> assert f(3) == func(3) == 5
    >>> assert f(3, 4) == func(3, 4) == 11
    >>> str(signature(f))
    '(x, y=1)'
    >>>
    >>> f = rm_params(func, params_to_remove='y z')
    >>> assert f(3) == func(3) == 5
    >>> str(signature(f))
    '(x)'

    But ``rm_params`` won't let you remove params that don't have defaults.

    >>> f = rm_params(func, params_to_remove='x z')
    Traceback (most recent call last):
    ...
    AssertionError: Some of the params you want to remove don't have defaults: {'x'}

    """
    if isinstance(params_to_remove, str):
        params_to_remove = params_to_remove.split()
    sig = Sig(func)
    params_to_remove_that_do_not_have_defaults = set(params_to_remove) & set(
        sig.without_defaults.names
    )
    assert not params_to_remove_that_do_not_have_defaults, (
        f"Some of the params you want to remove don't have defaults: "
        f'{params_to_remove_that_do_not_have_defaults}'
    )

    return wrap(
        func, ingress=include_exclude_ingress_factory(func, exclude=params_to_remove)
    )


#     new_sig = sig - params_to_remove
#     return new_sig(func)


def arg_val_converter(func, **conversion_for_arg):
    return Wrap(func, ingress=ArgValConverterIngress(func, **conversion_for_arg))


def arg_val_converter_ingress(func, __strict=True, **conversion_for_arg):
    sig = Sig(func)
    if __strict:
        conversion_names_that_are_not_func_args = conversion_for_arg.keys() - sig.names
        assert not conversion_names_that_are_not_func_args, (
            'Some of the arguments you want to convert are not argument names '
            f'for the function: {conversion_names_that_are_not_func_args}'
        )

    @sig
    def ingress(*args, **kwargs):
        # TODO: Make a helper function for this ak -> k -> proc -> ak pattern
        kwargs = sig.kwargs_from_args_and_kwargs(args, kwargs)
        kwargs = dict(convert_dict_values(kwargs, conversion_for_arg))
        args, kwargs = sig.args_and_kwargs_from_kwargs(kwargs)
        return args, kwargs

    return ingress


# TODO: Fits global pattern -- merge
class ArgValConverterIngress:
    def __init__(self, func, __strict=True, **conversion_for_arg):
        sig = Sig(func)
        if __strict:
            conversion_names_that_are_not_func_args = (
                conversion_for_arg.keys() - sig.names
            )
            assert not conversion_names_that_are_not_func_args, (
                'Some of the arguments you want to convert are not argument names '
                f'for the function: {conversion_names_that_are_not_func_args}'
            )
        self.sig = sig
        self.conversion_for_arg = conversion_for_arg
        wraps(func)(self)

    def __call__(self, *args, **kwargs):
        # TODO: Make a helper function for this ak -> k -> proc -> ak pattern
        kwargs = self.sig.kwargs_from_args_and_kwargs(args, kwargs)
        kwargs = dict(convert_dict_values(kwargs, self.conversion_for_arg))
        args, kwargs = self.sig.args_and_kwargs_from_kwargs(kwargs)
        return args, kwargs


# ---------------------------------------------------------------------------------------
# Utils to help define value conversions in ingress functions


def convert_dict_values(to_convert: dict, key_to_conversion_function: dict):
    for k, v in to_convert.items():
        if k in key_to_conversion_function:
            conversion_func = key_to_conversion_function[k]
            yield k, conversion_func(v)  # converted kv pair
        else:
            yield k, v  # unconverted kv pair


# TODO: Test for performance and ask about readability
def _alt_convert_dict_values(to_convert: dict, key_to_conversion_function: dict):
    for k, v in to_convert.items():
        conversion_func = key_to_conversion_function.get(k, lambda x: x)
        yield k, conversion_func(v)


def params_used_in_funcs(funcs):
    return {name for func in funcs for name in Sig(func).names}


def required_params_used_in_funcs(funcs):
    return {name for func in funcs for name in Sig(func).required_names}


from i2.signatures import _call_forgivingly


# TODO: Should we add some explicit/validation/strict options?
def kwargs_trans(
    kwargs: dict = None, /, _recursive=False, _inplace=False, **key_and_val_func
):
    """Transform a kwargs dict or build a transformer.

    :param kwargs: The dict containing the input kwargs that we will transform
    :param _recursive: Whether the transformations listed in ``key_and_val_func`` should
        be applied "recursively". When set to ``False`` (default), each transformation
        function applies to the original ``kwargs``, not the one that was transformed,
        so far.
    :param _inplace: If set to ``False`` will make a shallow copy of the ``kwargs``
        before transforming it (only relevant if ``_recursive=True``
    :param key_and_val_func: The ``key=val_func`` pairs that indicate that a
        ``val_func`` should be applied to the ``kwargs``, maching the argument names of
        the ``val_func`` to the keys of ``kwargs`` and extracting the values found
        therein to use for the corresponding inputs of that ``val_func``.
    :return: The transformed kwargs.

    >>> d = dict(a=1, b=2, c=3)
    >>> kwargs_trans(
    ...     d,
    ...     a=lambda a: a * 10,
    ...     b=lambda a, b: a + b
    ... )
    {'a': 10, 'b': 3, 'c': 3}

    See that ``d`` is unchanged here (transformation is not in place).

    >>> d
    {'a': 1, 'b': 2, 'c': 3}

    Typically you'll use ``kwargs_trans`` as a factory:

    >>> trans = kwargs_trans(a=lambda a: a * 10, b=lambda a, b: a + b)
    >>> trans(d)
    {'a': 10, 'b': 3, 'c': 3}

    Here we'll demo what the ``_recursive`` and ``_inplace`` arguments do.

    >>> from functools import partial
    >>> re_kwargs_trans = partial(kwargs_trans, _recursive=True, _inplace=True)
    >>> d = dict(a=1, b=2, c=3)
    >>>
    >>> re_kwargs_trans(
    ...     d,
    ...     a=lambda a: a * 10,
    ...     b=lambda a, b: a + b
    ...     # since _recursive=True, the a that is used is the new a = 10, not a = 1:
    ... )
    {'a': 10, 'b': 12, 'c': 3}

    Since ``_inplace=True``, ``d`` itself has changed:

    >>> d
    {'a': 10, 'b': 12, 'c': 3}

    Sometimes you'll pipe several transformers together:

    >>> from i2 import Pipe
    >>> trans = Pipe(
    ...     re_kwargs_trans(a=lambda a: a / 10),
    ...     re_kwargs_trans(c=lambda a,b,c: a * b * c),
    ...     # and then compute a new value of a using a and c:
    ...     re_kwargs_trans(a=lambda a, c: c - 1),
    ... )
    >>> trans(d)
    {'a': 35.0, 'b': 12, 'c': 36.0}
    """
    if kwargs is None:
        return partial(
            kwargs_trans, _recursive=_recursive, _inplace=_inplace, **key_and_val_func
        )
    if not _recursive:
        new_kwargs = dict()
        for key, val_func in key_and_val_func.items():
            new_kwargs[key] = _call_forgivingly(val_func, (), kwargs)
        return dict(kwargs, **new_kwargs)
    else:
        if _inplace is False:
            kwargs = kwargs.copy()
        for key, val_func in key_and_val_func.items():
            kwargs[key] = _call_forgivingly(val_func, (), kwargs)
        return kwargs


# ---------------------------------------------------------------------------------------

from inspect import Parameter
from dataclasses import make_dataclass

empty = Parameter.empty


def camelize(s):
    """
    >>> camelize('camel_case')
    'CamelCase'
    """
    return ''.join(ele.title() for ele in s.split('_'))


def kwargs_trans_to_extract_args_from_attrs(
    outer_kwargs: dict, attr_names=(), obj_param='self'
):
    self = outer_kwargs.pop(obj_param)
    arguments_extracted_from_obj = {name: getattr(self, name) for name in attr_names}
    # The kwargs we need are the union of the extracted arguments with the remaining outer_kwargs
    return dict(arguments_extracted_from_obj, **outer_kwargs)


# TODO: kind lost here, only 3.10 offers dataclasses with some control over kind:
#   See: https://stackoverflow.com/questions/49908182/how-to-make-keyword-only-fields-with-dataclasses
def param_to_dataclass_field_tuple(param: Parameter):
    return param.name, param.annotation, param.default


MethodFunc = Callable


def func_to_method_func(
    func,
    instance_params=(),
    *,
    method_name=None,
    method_params=None,
    instance_arg_name='self',
) -> MethodFunc:
    """Get a 'method function' from a 'normal function'. Also known as "methodize".

    That is, get a function that gives the same outputs as the 'normal function',
    except that some of the arguments are sourced from the attributes of the first
    argument.

    The intended use case is when you want to inject one or several methods in a class
    or instance, sourcing some of the arguments of the underlying function from a
    common pool: The attributes of the instance.

    Consider the following function involving four parameters: ``a, b, c`` and ``d``.

    >>> def func(a, b: int, c=2, *, d='bar'):
    ...     return f"{d}: {(a + b) * c}"
    >>> func(1, 2, c=3, d='hello')
    'hello: 9'

    If we wanted to make an equivalent "method function" that would source it's ``a``
    and it's ``c`` from the first argument's (in practice this first argument will be
    and instance of the class the method will be bound to), we can do so like so:

    >>> method_func = func_to_method_func(func, 'a c')
    >>> from inspect import signature
    >>> str(signature(method_func))
    "(self, b: int, *, d='bar')"

    Note that the first argument is ``self`` (default name for an "instance"),
    that ``a`` and ``c`` are not there, but that the two remaining parameters,
    ``b`` and ``d`` are present, in the same order, and with the same annotations and
    parameter kind (the ``d`` is still keyword-only).

    Now let's make a dummy object that has attributes ``a`` and a ``c``, and use it to
    call ``method_func``:

    >>> from collections import namedtuple
    >>> instance = namedtuple('FakeInstance', 'a c')(1, 3)
    >>> method_func(instance, 2, d='hello')
    'hello: 9'

    Which is:

    >>> assert method_func(instance, 2, d='hello') == func(1, 2, c=3, d='hello')

    Often, though, what you'll want is to include this method function directly in
    a class, as you're making that class "normally". That works too:

    >>> from dataclasses import dataclass
    >>> @dataclass
    ... class Klass:
    ...     a : int = 1
    ...     c : int = 3
    ...     method_func = func_to_method_func(func, 'a c')
    >>> instance = Klass(1, 3)
    >>> instance.method_func(2, d='hello')
    'hello: 9'

    What if your function has argument names that don't correspond to the names you
    have, or want, as attributes of the class? Or even, you have several functions that
    share an argument name that need to be bound to a different attribute?

    For that, just use ``map_names`` to wrap the function, giving it the names that
    you need to give it to have the effect you want (the binding of those arguments
    to attributes of the instance):

    >>> from i2.wrapper import ch_names
    >>> def func(x, y: int, z=2, *, d='bar'):
    ...     return f"{d}: {(x + y) * z}"
    >>> from dataclasses import dataclass
    >>> @dataclass
    ... class Klass:
    ...     a : int = 1
    ...     c : int = 3
    ...     method_func = func_to_method_func(ch_names(func, x='a', z='c'), 'a c')
    >>> instance = Klass(1, 3)
    >>> instance.method_func(2, d='hello')
    'hello: 9'

    """
    # get a signature object for func
    sig = Sig(func)
    # if method_name not give, use the function's name
    method_name = method_name or sig.name
    # if params expressed as string, split it into a list of parameter (names)
    if isinstance(instance_params, str):
        instance_params = instance_params.split()
    if isinstance(method_params, str):
        method_params = method_params.split()
    # if method_params is not given, take those parameters that aren't in instance_params
    method_params = method_params or tuple(
        name for name in sig.names if name not in instance_params
    )
    # the Sig object of the method: instance name followed with method_params
    method_sig = instance_arg_name + Sig(func)[method_params]
    # make the ingress function that will map method_sig's interface to sig's.
    ingress = Ingress(
        inner_sig=sig,
        # inside, foo will be doing the work, so need to map to its signature
        kwargs_trans=partial(
            # this is how to transform outer (args, kwargs) to inner ones
            kwargs_trans_to_extract_args_from_attrs,
            attr_names=instance_params,
            obj_param=instance_arg_name,
        ),
        outer_sig=method_sig,  # this is the signature we want at the interface
    )
    # wrap the function, name it and return it
    method_func = wrap(func, ingress=ingress)
    method_func.__name__ = method_name
    return method_func


# TODO: See bind_funcs_object_attrs_old and get rid of it if bind_funcs_object_attrs
#  is better and has all old has.
# TODO: Make instances picklable (class is already, but not instances!)!!
def bind_funcs_object_attrs(
    funcs, init_params=(), *, cls=None, module=None, **extra_attrs
):
    """Transform one or several functions into a class that contains them as methods
    sourcing specific arguments from the instance's attributes.

    >>> from inspect import signature
    >>> from dataclasses import dataclass
    >>>
    >>> def foo(a, b, c=2, *, d='bar'):
    ...     return f"{d}: {(a + b) * c}"
    >>> foo(1, 2)
    'bar: 6'
    >>> Klass = bind_funcs_object_attrs(foo, init_params='a c')
    >>> Klass.__name__
    'Foo'
    >>> instance = Klass(a=1, c=3)
    >>> assert instance.foo(2, d='hello') == 'hello: 9' == foo(
    ...     a=1, b=2, c=3, d='hello')
    >>> str(signature(Klass))
    '(a, c=2) -> None'
    >>>
    >>> instance = Klass(a=1, c=3)
    >>> str(instance)
    'Foo(a=1, c=3)'
    >>> str(signature(instance.foo))
    "(b, *, d='bar')"
    >>> instance.foo(2, d='hello')
    'hello: 9'
    >>> instance.foo(10, d='goodbye')
    'goodbye: 33'

    >>> def foo(a, b, c):
    ...     return a + b * c
    ...
    >>> def bar(d, e):
    ...     return f"{d=}, {e=}"
    ...
    >>> @dataclass
    ... class K:
    ...     a: int
    ...     e: int
    ...
    >>> C = bind_funcs_object_attrs([foo, bar], 'a e', cls=K)
    >>> str(signature(C))
    '(a: int, e: int) -> None'
    >>> c = C(1,2)
    >>> assert str(signature(c.foo)) == '(b, c)'
    >>> c.foo(3,4)
    13
    >>> assert str(signature(c.bar)) == '(d)'
    >>> c.bar(5)
    'd=5, e=2'
    """

    if isinstance(init_params, str):
        init_params = init_params.split()

    dflt_cls_name = 'FuncsUnion'
    if callable(funcs) and not isinstance(funcs, Iterable):
        single_func = funcs
        dflt_cls_name = camelize(getattr(single_func, '__name__', dflt_cls_name))
        funcs = [single_func]

    # If cls not given, make one from the funcs and init_params
    if not isinstance(cls, type):
        # if the class is not given, we need to make one
        # ... for this we make a name
        if isinstance(cls, str):
            cls_name = cls
        else:
            cls_name = dflt_cls_name
        # ... and an actual class
        cls = _mk_base_class_for_funcs(cls_name, init_params, funcs)

    for func in funcs:
        method_func = func_to_method_func(func, init_params)
        setattr(cls, method_func.__name__, method_func)

    if module:
        cls.__module__ = module
    for attr_name, attr_val in extra_attrs.items():
        setattr(cls, attr_name, attr_val)

    return cls


class PickleHelperMixin:
    def __reduce__(self):
        return self.__name__


def _mk_base_class_for_funcs(cls_name, init_params, funcs):
    """
    Make a class with given init_params.

    :param cls_name: The name the class should have
    :param init_params: list of strings used to determine what arg names are in the init
    :param funcs: list of functions used to add defaults and annotations to the init args
    :return:
    """
    # Make the signature for the init
    class_init_sig = Sig(init_params)
    if funcs:
        for func in funcs:
            init_params_in_func_sig = list(
                filter(None, map(Sig(func).get, init_params))
            )
            class_init_sig = class_init_sig.merge_with_sig(
                init_params_in_func_sig,
                default_conflict_method='fill_defaults_and_annotations',
            )

    dataclass_fields = list(map(param_to_dataclass_field_tuple, class_init_sig.params))
    cls = make_dataclass(cls_name, dataclass_fields, bases=(PickleHelperMixin,))

    return cls


def bind_funcs_object_attrs_old(
    funcs, init_params=(), *, cls=None,
):
    """Transform one or several functions into a class that contains them as methods
    sourcing specific arguments from the instance's attributes.
    >>> from inspect import signature
    >>> from dataclasses import dataclass
    >>>
    >>> def foo(a, b, c=2, *, d='bar'):
    ...     return f"{d}: {(a + b) * c}"
    >>> foo(1, 2)
    'bar: 6'
    >>> Klass = bind_funcs_object_attrs_old(foo, init_params='a c')
    >>> Klass.__name__
    'Foo'
    >>> instance = Klass(a=1, c=3)
    >>> assert instance.foo(2, d='hello') == 'hello: 9' == foo(
    ...     a=1, b=2, c=3, d='hello')
    >>> str(signature(Klass))
    '(a, c=2) -> None'
    >>>
    >>> instance = Klass(a=1, c=3)
    >>> str(instance)
    'Foo(a=1, c=3)'
    >>> str(signature(instance.foo))
    "(b, *, d='bar')"
    >>> instance.foo(2, d='hello')
    'hello: 9'
    >>> instance.foo(10, d='goodbye')
    'goodbye: 33'
    >>> def foo(a, b, c):
    ...     return a + b * c
    ...
    >>> def bar(d, e):
    ...     return f"{d=}, {e=}"
    ...
    >>> @dataclass
    ... class K:
    ...     a: int
    ...     e: int
    ...
    >>> C = bind_funcs_object_attrs([foo, bar], 'a e', cls=K)
    >>> str(signature(C))
    '(a: int, e: int) -> None'
    >>> c = C(1,2)
    >>> assert str(signature(c.foo)) == '(b, c)'
    >>> c.foo(3,4)
    13
    >>> assert str(signature(c.bar)) == '(d)'
    >>> c.bar(5)
    'd=5, e=2'
    """

    if isinstance(init_params, str):
        init_params = init_params.split()

    dflt_cls_name = 'FuncsUnion'
    if callable(funcs) and not isinstance(funcs, Iterable):
        single_func = funcs
        dflt_cls_name = camelize(getattr(single_func, '__name__', dflt_cls_name))
        funcs = [single_func]

    if not isinstance(cls, type):
        # if the class is not given, we need to make one
        if isinstance(cls, str):
            cls_name = cls
        else:
            cls_name = dflt_cls_name

        # init_parameter_objects = Sig(func)[init_params].params
        # Make the signature for the init
        class_init_sig = Sig()
        for func in funcs:
            class_init_sig = class_init_sig.merge_with_sig(func)[init_params]

        dataclass_fields = list(
            map(param_to_dataclass_field_tuple, class_init_sig.params)
        )
        cls = make_dataclass(cls_name, dataclass_fields)

    for func in funcs:
        method_func = func_to_method_func(func, init_params)
        setattr(cls, method_func.__name__, method_func)
    return cls


def _items_filt(d: dict, keys):
    for k, v in d.items():
        if k in keys:
            yield k, v


def _mk_sig_from_params_and_funcs(params, funcs):
    def gen():
        for param in params:
            pass


## An attempt to redo a func_to_method_func because I forgot func_to_method_func existed
## Has some different ideas, so keeping around until I decide it's time to let go.
# NoSuchKey = type('NoSuchKey', (), {})
# _instance_extractor: KwargsTrans
#
# # TODO: Add more (possibly optional) bind validation to fail early.
# def _instance_extractor(
#     outer_kwargs, bind: IdentifierMapping = (), instance_param: Identifier = 'self'
# ):
#     """
#
#     :param outer_kwargs: The input/outer keyword arguments
#     :param bind: The inner->outer param mapping that defines what we want to extract
#     :param instance_param: The name of the outer_kwargs key that contains the 'instance'
#         from which we'll extract the bound
#     :return:
#     """
#     """A KwargsTrans that extracts need arguments from one of the 'instance'
#     outer_kwargs values.
#
#
#     """
#     # Compute the inverse {outer:inner,...} of {inner:outer,...} bind
#     inv_bind = invert_map(bind)
#     outer_kwargs = outer_kwargs.copy()
#     instance = outer_kwargs.pop(instance_param)  # TODO: Better error handing
#
#     def gen():
#         for outer_param, outer_val in outer_kwargs.items():
#             if inner_param := inv_bind.get(outer_param, NoSuchKey) is not NoSuchKey:
#                 # if outer_param was bound, the bound inner_param should be tied to
#                 # the instance's attribute
#                 yield inner_param, getattr(instance, outer_val)
#             else:
#                 # take the arg name and val as is
#                 yield outer_param, outer_val
#
#     return dict(gen())

#
# def methodize(func, bind: Bind = ()):
#     bind = identifier_mapping(bind)
#     ingress = Ingress(
#         outer_sig=Sig(func),
#         kwargs_trans=partial(_instance_extractor, bind=bind),
#         inner_sig=Sig('self') + Sig(func) - Sig(list(bind)),  # TODO: solve type or lint
#     )
#     return wrap(func, ingress=ingress)


# ---------------------------------------------------------------------------------------
# Extended Wrapper class

from i2 import call_forgivingly


class WrapperValidationError(ValueError):
    """Raised when wrapper some construction params are not valid"""


class EgressValidationError(WrapperValidationError):
    """Raised when a egress is not valid"""


class IngressValidationError(WrapperValidationError):
    """Raised when a ingress is not valid"""


class CallerValidationError(WrapperValidationError):
    """Raised when a caller is not valid"""


def _default_ingress(*args, **kwargs):
    return args, kwargs


def _default_egress(output, **egress_params):
    return output


def _default_caller(func, args, kwargs):
    return func(*args, **kwargs)


_keyword_kinds = {Sig.KEYWORD_ONLY, Sig.VAR_KEYWORD}


def _all_kinds_are_keyword_only_or_variadic_keyword(sig):
    return all(kind in _keyword_kinds for kind in list(sig.kinds.values())[3:])


# TODO: Factor out more common parts with Wrap and reuse (possibly through _Wrap)
class Wrapx(_Wrap):
    def __init__(self, func, ingress=None, egress=None, *, caller=None, name=None):
        """An extended wrapping object that allows more complex wrapping mechanisms.

        :param func: The wrapped function
        :param ingress: The incoming data transformer. It determines the argument properties
            (name, kind, default and annotation) as well as the actual input of the
            wrapped function.
        :param egress: The outgoing data transformer. It also takes precedence over the
            wrapped function to determine the return annotation of the ``Wrap`` instance
        :param caller: A caller defines what it means to call the ``func`` on the
            arguments it is given. It should be of the form
            ``caller(func, args, kwargs, *, ...extra_keyword_only_params)``.
            By default, the caller will simply return ``func(*args, **kwargs)``.
        :param name: Name to give the wrapper (will use wrapped func name by default)

        :return: A callable instance wrapping ``func``

        >>> from inspect import signature
        >>>
        >>> def func(x, y):
        ...     return x + y
        ...
        >>> def save_on_output_egress(v, *, k, s):
        ...     s[k] = v
        ...     return v
        ...
        >>> save_on_output = Wrapx(func, egress=save_on_output_egress)
        >>> # TODO: should be `(x, y, *, k, s)` --> Need to work on the merge for this.
        >>> str(signature(save_on_output))
        '(x, y, k, s)'
        >>>
        >>> store = dict()
        >>> save_on_output(1, 2, k='save_here', s=store)
        3
        >>> assert save_on_output(1, 2, k='save_here', s=store) == 3 == func(1, 2)
        >>> store  # see what's in the store now!
        {'save_here': 3}

        A caller is meant to control the way the function is called.
        It is given the ``func`` and the ``func_args`` and ``func_kwargs``
        (whatever the ingress function gives it, if present) and possibly additional
        params and will return... well, what ever you tell it to.

        This can be used, for example, to call the function in a subprocess,
        or on a remote system, differ computation (command pattern, for example, using
        ``functools.partial``, or do what ever needs to have a view both on the function
        and its inputs.

        Here, we will wrap the function so it will apply to an iterable of inputs
        (of the first argument), returning a list of results

        >>> def func(x, y=2):
        ...     return x + y
        ...
        >>> def iterize(func, args, kwargs):
        ...     first_arg_val = next(iter(kwargs.values()))
        ...     return list(map(func, first_arg_val))
        ...
        >>> iterized_func = Wrapx(func, caller=iterize)
        >>> iterized_func([1, 2, 3, 4])
        [3, 4, 5, 6]

        Let's do the same as above, but allow other variables (here ``y``) to be input as
        well. This takes a bit more work...

        >>> from functools import partial
        >>> def _iterize_first_arg(func, args, kwargs):
        ...     first_arg_name = next(iter(kwargs))
        ...     remaining_kwargs = {
        ...         k: v for k, v in kwargs.items() if k != first_arg_name
        ...     }
        ...     return list(
        ...         map(partial(func, **remaining_kwargs), kwargs[first_arg_name])
        ...     )

        Let's demo a different way of using Wrapx: Making a wrapper to apply at
        function definition time

        >>> iterize_first_arg = partial(Wrapx, caller=_iterize_first_arg)
        >>> @iterize_first_arg
        ... def func(x, y):
        ...     return x + y
        >>>
        >>> func([1, 2, 3, 4], 10)
        [11, 12, 13, 14]


        """
        super().__init__(func, ingress, egress, caller=caller, name=name)
        self.ingress, self.egress, self.caller, self.sig = _process_wrapx_params(
            func, ingress, egress, caller
        )

        self.__signature__ = self.sig
        self.__wrapped__ = func
        # TODO: Pros and cons analysis of pointing __wrapped__ to func. partial uses
        #  .func, but wraps looks for __wrapped__

    def __call__(self, *args, **kwargs):
        try:
            _kwargs = self.sig.kwargs_from_args_and_kwargs(args, kwargs)
            # TODO: Consider call_forgivingly(self.ingress, *args, **kwargs)
            #  because call_forgivingly(self.ingress, **_kwargs) doesn't allow
            #  same ingress functions to be used for Wrap and Wrapx
            func_args, func_kwargs = call_forgivingly(self.ingress, **_kwargs)
            inner_output = call_forgivingly(
                self.caller, self.func, func_args, func_kwargs
            )
            return call_forgivingly(self.egress, inner_output, **_kwargs)
        except Exception as e:
            # Try running again, but with more careful validation, to try to give
            # more specific error messages
            # We don't do this in the first run so that we don't incur the validation
            # overhead on every call.
            _kwargs = self.sig.kwargs_from_args_and_kwargs(args, kwargs)
            func_args, func_kwargs = _validate_ingress_output(
                call_forgivingly(self.ingress, **_kwargs)
            )
            inner_output = call_forgivingly(self.func, *func_args, **func_kwargs)
            return call_forgivingly(self.egress, inner_output, **_kwargs)


def _validate_ingress_output(ingress_output):
    if (
        not isinstance(ingress_output, tuple)
        or not len(ingress_output) == 2
        or not isinstance(ingress_output[0], tuple)
        or not isinstance(ingress_output[1], dict)
    ):
        raise IngressValidationError(
            f'An ingress function should return a (tuple, dict) pair. '
            f'This ingress function returned: {ingress_output}'
        )
    return ingress_output


def _process_wrapx_params(func, ingress, egress, caller):
    func_sig = Sig(func)

    ingress, ingress_sig = _init_ingress(func_sig, ingress)
    egress, egress_sig = _init_egress(func_sig, egress)
    caller, caller_sig = _init_caller(caller)

    sig = _mk_composite_sig(ingress_sig, egress_sig, caller_sig)
    return ingress, egress, caller, sig


def _mk_composite_sig(ingress_sig, egress_sig, caller_sig):
    egress_sig_minus_first_arg = egress_sig - egress_sig.names[0]
    caller_sig_minus_three_first_args = caller_sig - caller_sig.names[:3]
    sig = Sig(
        ingress_sig + egress_sig_minus_first_arg + caller_sig_minus_three_first_args,
        return_annotation=egress_sig.return_annotation,
    )
    return sig


def _init_caller(caller):
    if caller is None:
        caller = _default_caller
        caller_sig = Sig('func args kwargs')  # sig with three inputs
    else:
        caller_sig = Sig(caller)
        if len(caller_sig) < 3:
            raise CallerValidationError(
                f'A caller must have at least three arguments: '
                f'{caller} signature was {caller_sig}'
            )
        if not _all_kinds_are_keyword_only_or_variadic_keyword(caller_sig):
            raise CallerValidationError(
                f'A caller must have at least three arguments'
                f'{caller} signature was {caller_sig}'
            )
    return caller, caller_sig


def _init_egress(func_sig, egress):
    if egress is None:
        egress = _default_egress
        # signature with a single 'output' arg and func_sig's return_annotation
        egress_sig = Sig('output', return_annotation=func_sig.return_annotation)
        return_annotation = func_sig.return_annotation
    else:
        egress_sig = Sig(
            egress,
            # if egress has no return_annotation, use the func_sig's one.
            # TODO: Is this really correct/safe? What if egress changes the type?
            return_annotation=Sig(egress).return_annotation
            or func_sig.return_annotation,
        )

    return egress, egress_sig


def _init_ingress(func_sig, ingress):
    if ingress is None:
        ingress = _default_ingress
        ingress_sig = func_sig
    else:
        ingress_sig = Sig(ingress)
    return ingress, ingress_sig


# ---------------------------------------------------------------------------------------
# partialx


def partialx(
    func, *args, __name__=None, _rm_partialize=False, _allow_reordering=False, **kwargs
):
    """
    Extends the functionality of builtin ``functools.partial`` with the ability to

    - set ``__name__ ``

    - remove partialized arguments from signature

    - reorder params (so that defaults are at the end)

    >>> def f(a, b=2, c=3):
    ...     return a + b * c
    >>> curried_f = partialx(f, c=10, _rm_partialize=True)
    >>> curried_f.__name__
    'f'
    >>> from inspect import signature
    >>> str(signature(curried_f))
    '(a, b=2)'

    >>> def f(a, b, c=3):
    ...     return a + b * c

    Note that ``a`` gets a default, but ``b`` does not, yet is after ``a``.
    This is allowed because these parameters all became KEYWORD_ONLY.

    >>> g = partialx(f, a=1)
    >>> str(Sig(g))
    '(*, a=1, b, c=3)'

    If you wanted to reorder the parameters to have all defaulted kinds be at the end,
    as usual, you can do so using ``_allow_reordering=True``

    >>> g = partialx(f, a=1, _allow_reordering=True)
    >>> str(Sig(g))
    '(*, b, a=1, c=3)'

    """
    f = partial(func, *args, **kwargs)
    if _rm_partialize:
        sig = Sig(func)
        partialized = list(
            sig.kwargs_from_args_and_kwargs(args, kwargs, allow_partial=True)
        )
        sig = sig - partialized
        f = sig(partial(f, *args, **kwargs))
    if _allow_reordering:
        # TODO: Instead of Sig(f).defaults, move only params that need to move
        # TODO: + Change signature so that the number of params that become keyword-only
        #   is minimize.
        f = move_params_to_the_end(f, Sig(f).defaults)
    f.__name__ = __name__ or name_of_obj(func)
    return f


def move_params_to_the_end(
    func: Callable, names_to_move: Union[Callable, Iterable[str]]
):
    """
    Choose args from func, according to choice_args_func and move them
    to the right

    >>> from functools import partial
    >>> from i2 import Sig
    >>> def foo(a, b, c):
    ...     return a + b + c
    >>> g = partial(foo, b=4)  # fixing a, which is before b
    >>> h = move_params_to_the_end(g, Sig(g).defaults)
    >>> assert str(Sig(g)) == '(a, *, b=4, c)'
    >>> assert str(Sig(h)) == '(a, *, c, b=4)'

    """
    if callable(names_to_move):
        names_to_move = names_to_move(func)
    assert isinstance(names_to_move, Iterable), (
        f'names_to_move must be an iterable of names '
        f'or a callable producing one from a function. Was {names_to_move}'
    )

    names = Sig(func).names
    reordered = move_names_to_the_end(names, names_to_move)
    wrapped_func = include_exclude(func, include=reordered)
    return wrapped_func


def move_names_to_the_end(names, names_to_move_to_the_end):
    """
    Remove the items of ``names_to_move_to_the_end`` from ``names``
    and append to the right of names

    >>> names = ['a','c','d','e']
    >>> names_to_move_to_the_end = ['c','e']
    >>> move_names_to_the_end(names, names_to_move_to_the_end)
    ['a', 'd', 'c', 'e']
    >>> names_to_move_to_the_end = 'c e'
    >>> move_names_to_the_end(names, names_to_move_to_the_end)
    ['a', 'd', 'c', 'e']

    """
    if isinstance(names_to_move_to_the_end, str):
        names_to_move_to_the_end = names_to_move_to_the_end.split()
    else:
        names_to_move_to_the_end = list(names_to_move_to_the_end)
    removed = [x for x in names if x not in names_to_move_to_the_end]
    return list(removed) + names_to_move_to_the_end



"""Merge mappings

Marked for deprecation.
"""

from collections import ChainMap

from collections.abc import Mapping
from itertools import chain

is_mapping = lambda x: isinstance(x, Mapping)
not_mapping = lambda x: not (is_mapping(x))

from collections.abc import Iterable


def is_iterable(x):
    """Similar in nature to :func:`callable`, ``is_iterable`` returns
    ``True`` if an object is `iterable`_, ``False`` if not.

    >>> is_iterable([])
    True
    >>> is_iterable(1)
    False"""
    return isinstance(x, Iterable)


# Note: Mark for moving --> used in only one known place
def unique_iter(src, key=None):
    """Yield unique elements from the iterable, *src*, based on *key*,
    in the order in which they first appeared in *src*.

    >>> repetitious = [1, 2, 3] * 10
    >>> list(unique_iter(repetitious))
    [1, 2, 3]

    By default, *key* is the object itself, but *key* can either be a
    callable or, for convenience, a string name of the attribute on
    which to uniqueify objects, falling back on identity when the
    attribute is not present.

    >>> pleasantries = ['hi', 'hello', 'ok', 'bye', 'yes']
    >>> list(unique_iter(pleasantries, key=lambda x: len(x)))
    ['hi', 'hello', 'bye']
    """
    if not is_iterable(src):
        raise TypeError('expected an iterable, not %r' % type(src))
    if key is None:
        key_func = lambda x: x
    elif callable(key):
        key_func = key
    elif isinstance(key, str):
        key_func = lambda x: getattr(x, key, x)
    else:
        raise TypeError('"key" expected a string or callable, not %r' % key)
    seen = set()
    for i in src:
        k = key_func(i)
        if k not in seen:
            seen.add(k)
            yield i
    return


# Note: Mark for moving to py2store --> used in only one place (py2dash).
class ChainMapTree(Mapping):
    """Combine/overlay multiple hierarchical mappings. This efficiently merges
    multiple hierarchical (could be several layers deep) dictionaries, producing
    a new view into them that acts exactly like a merged dictionary, but without
    doing any copying.
    Because it doesn't actually copy the data, it is intended to be used only
    with immutable mappings. It is safe to change *leaf* data values,
    and the results will be reflected here, but changing the structure of any
    of the trees will not work.

    >>> base1 = {
    ...     'a1': 'base1.a1',
    ...     'a2': 'base1.a2',
    ...     'a3': {
    ...         'b1': 'base1.a3.b1',
    ...         'b2': 'base1.a3.b2',
    ...     },
    ... }
    >>> base2 = {
    ...     'a2': 'base2.a2',
    ...     'a3': {
    ...         'b2': 'base2.a3.b2',
    ...         'b4': 'base2.a3.b4',
    ...     },
    ...     'a4': 'base2.a4',
    ... }
    >>>
    >>> cm = ChainMapTree(base1, base2)
    >>> cm['a1']
    'base1.a1'
    >>> cm['a2']
    'base1.a2'
    >>> cm['a4']
    'base2.a4'
    >>> cm['a3']
    ChainMapTree({'b1': 'base1.a3.b1', 'b2': 'base1.a3.b2'}, {'b2': 'base2.a3.b2', 'b4': 'base2.a3.b4'})
    >>> cm['a3']['b1']
    'base1.a3.b1'
    >>> cm['a3']['b4']
    'base2.a3.b4'
    >>> cm = ChainMapTree(base2, base1)
    >>> cm['a1']
    'base1.a1'
    >>> cm['a2']
    'base2.a2'
    >>> cm['a4']
    'base2.a4'
    >>> cm['a3']
    ChainMapTree({'b2': 'base2.a3.b2', 'b4': 'base2.a3.b4'}, {'b1': 'base1.a3.b1', 'b2': 'base1.a3.b2'})
    >>> cm['a3']['b2']
    'base2.a3.b2'
    >>> cm['a3']['b1']
    'base1.a3.b1'
    >>>
    >>> # Let's do a ChainMapTree with THREE bases now!
    >>> base3 = {
    ...     'a2': 'base3.a2',
    ...     'a3': {
    ...         'b2': 'base3.a3.b2',
    ...         'b4': 'base3.a3.b4',
    ...     },
    ...     'a4': 'base3.a4',
    ... }
    >>> cm = ChainMapTree(base3, base2, base1)
    >>> cm['a2']  # will get it from base3
    'base3.a2'
    >>> cm['a3']['b2']  # will get it from base3 (not base2)
    'base3.a3.b2'
    >>> cm['a3']['b1']  # will get it from base1 (since no one else has it)
    'base1.a3.b1'

    Based on: https://gist.github.com/Klortho/7d83975559bdcc47ac64fd7d877934f6
    """

    _max_repr_keys = (
        7  # Not used (anymore) at this point. Was to control __repr__ output length
    )

    def __init__(self, *maps):
        _maps = list(maps)

        # All keys of kids that are mappings
        kid_keys = set([key for m in maps for key in m.keys() if is_mapping(m[key])])

        # This will be a dictionary of lists of mappings
        kid_maps = {}
        for key in kid_keys:
            # The list of child mappings for this key
            kmaps = [m[key] for m in maps if key in m]
            # Make sure they are *all* mappings
            if any(map(not_mapping, kmaps)):
                raise KeyError(key)
            # Recurse
            kid_maps[key] = ChainMapTree(*kmaps)

        # If non-empty, prepend it to the existing list
        if len(kid_maps.keys()) > 0:
            _maps.insert(0, kid_maps)

        self._maps = _maps

    def __getitem__(self, key):
        for _map in self._maps:
            try:
                return _map[key]
            except KeyError:
                pass
        raise KeyError(key)

    def __iter__(self):
        return unique_iter(chain(*(_map.__iter__() for _map in self._maps)))

    def __len__(self):
        return len(list(self.__iter__()))

    def _mk_keys_str(self, keys):
        first_keys = list()
        i = -1
        for i, k in enumerate(keys):
            if i < self._max_repr_keys:
                first_keys.append(k)
            else:
                break
        keys_str = ', '.join(first_keys)
        if i >= self._max_repr_keys:
            keys_str += ', ...'
        return keys_str

    def to_dict(self):
        """Convert to dict

        >>> a = {'a': {'x': 1, 'z': 3}, 'foo': "a's foo"}
        >>> b = {'a': {'y': 222, 'z': 333}, 'foo': "b's foo"}
        >>> cm = ChainMapTree(a, b)
        >>> # It acts like a dict when you ask for items, but is not a dict. If you want a dict, do this:
        >>> cm.to_dict()
        {'a': {'x': 1, 'z': 3, 'y': 222}, 'foo': "a's foo"}
        >>> # Compare to normal/flat/not-nested chaining:
        >>> dict(a, **b)   # Note the precedence is the inverse of ChainMapTree!
        {'a': {'y': 222, 'z': 333}, 'foo': "b's foo"}
        >>>
        >>> # See what you get if you specify b before a
        >>> ChainMapTree(b, a).to_dict()
        {'a': {'y': 222, 'z': 333, 'x': 1}, 'foo': "b's foo"}
        >>> # Compare to normal/flat/not-nested chaining:
        >>> dict(b, **a)  # Note the precedence is the inverse of ChainMapTree!
        {'a': {'x': 1, 'z': 3}, 'foo': "a's foo"}
        """
        d = dict()
        for k in self:
            v = self[k]
            if not isinstance(v, ChainMapTree):
                d[k] = v
            else:
                d[k] = v.to_dict()
        return d

    def __repr__(self):
        return f"{self.__class__.__name__}({', '.join((_map.__repr__() for _map in self._maps))})"
        #         map_keys_strs = self._mk_keys_str(self.keys())
        #         map_keys_strs = []
        #         for _map in self._maps:
        #             map_keys_strs.append(self._mk_keys_str(_map.keys()))
        #         map_keys_strs = ', '.join(map('({})'.format, map_keys_strs))
        # return f"{self.__class__.__name__}({map_keys_strs})"
